import json
from unittest.mock import MagicMock

import pytest
from event_model import Event

from mx_bluesky.beamlines.i04.callbacks.murko_callback import MurkoCallback

test_oav_uuid = "UUID"
test_smargon_data = 90


def event_template(data_key, data_value) -> Event:
    return {
        "descriptor": "bd45c2e5-2b85-4280-95d7-a9a15800a78b",
        "time": 1666604299.828203,
        "data": {data_key: data_value},
        "timestamps": {data_key: 1666604299.8220396},
        "seq_num": 1,
        "uid": "29033ecf-e052-43dd-98af-c7cdd62e8173",
        "filled": {},
    }


test_oav_event = event_template("oav_to_redis_forwarder-uuid", test_oav_uuid)
test_smargon_event = event_template("smargon-omega", test_smargon_data)

test_start_document = {
    "uid": "event_uuid",
    "zoom_percentage": 80,
    "microns_per_x_pixel": 1.2,
    "microns_per_y_pixel": 2.5,
    "beam_centre_i": 158,
    "beam_centre_j": 452,
    "sample_id": 12345,
}


@pytest.fixture
def murko_callback() -> MurkoCallback:
    callback = MurkoCallback("", "")
    callback.redis_client = MagicMock()
    return callback


@pytest.fixture
def murko_with_mock_call(murko_callback) -> MurkoCallback:
    murko_callback.call_murko = MagicMock()
    return murko_callback


def test_when_oav_data_arrives_then_murko_not_called(
    murko_with_mock_call: MurkoCallback,
):
    murko_with_mock_call.event(test_oav_event)
    murko_with_mock_call.call_murko.assert_not_called()  # type: ignore


def test_when_smargon_data_arrives_then_murko_not_called(
    murko_with_mock_call: MurkoCallback,
):
    murko_with_mock_call.event(test_smargon_event)
    murko_with_mock_call.call_murko.assert_not_called()  # type: ignore


def test_when_smargon_data_arrives_before_oav_data_then_murko_not_called(
    murko_with_mock_call: MurkoCallback,
):
    murko_with_mock_call.event(test_smargon_event)
    murko_with_mock_call.event(test_oav_event)
    murko_with_mock_call.call_murko.assert_not_called()  # type: ignore


def test_when_smargon_data_arrives_before_oav_data_then_murko_called_with_smargon_data(
    murko_with_mock_call: MurkoCallback,
):
    murko_with_mock_call.event(test_oav_event)
    murko_with_mock_call.event(test_smargon_event)
    murko_with_mock_call.call_murko.assert_called_once_with(  # type: ignore
        test_oav_uuid, test_smargon_data
    )


def test_when_murko_called_with_event_data_then_meta_data_put_in_redis(
    murko_callback: MurkoCallback,
):
    murko_callback.start(test_start_document)  # type: ignore
    murko_callback.event(test_oav_event)
    murko_callback.event(test_smargon_event)

    expected_metadata = {
        "zoom_percentage": 80,
        "microns_per_x_pixel": 1.2,
        "microns_per_y_pixel": 2.5,
        "beam_centre_i": 158,
        "beam_centre_j": 452,
        "sample_id": 12345,
        "omega_angle": test_smargon_data,
        "uuid": test_oav_uuid,
    }

    murko_callback.redis_client.hset.assert_called_once_with(  # type: ignore
        "murko:12345:metadata", test_oav_uuid, json.dumps(expected_metadata)
    )
    murko_callback.redis_client.publish.assert_called_once_with(  # type: ignore
        "murko", json.dumps(expected_metadata)
    )



from unittest.mock import ANY, MagicMock, call, patch

import bluesky.plan_stubs as bps
import pytest
from dodal.beamlines.i24 import I24_ZEBRA_MAPPING
from ophyd_async.testing import get_mock_put, set_mock_value

from mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2 import (
    collection_aborted_plan,
    collection_complete_plan,
    enter_hutch,
    initialise_extruder,
    laser_check,
    main_extruder_plan,
    read_parameters,
    tidy_up_at_collection_end_plan,
)
from mx_bluesky.beamlines.i24.serial.parameters import BeamSettings, ExtruderParameters
from mx_bluesky.beamlines.i24.serial.setup_beamline import Eiger, Pilatus

from ..conftest import TEST_LUT


@pytest.fixture
def dummy_params():
    params = {
        "visit": "/tmp/dls/i24/extruder/foo",
        "directory": "bar",
        "filename": "protein",
        "exposure_time_s": 0.1,
        "detector_distance_mm": 100,
        "detector_name": "eiger",
        "transmission": 1.0,
        "num_images": 10,
        "pump_status": False,
    }
    return ExtruderParameters(**params)


@pytest.fixture
def dummy_params_pp():
    params_pp = {
        "visit": "/tmp/dls/i24/extruder/foo",
        "directory": "bar",
        "filename": "protein",
        "exposure_time_s": 0.1,
        "detector_distance_mm": 100,
        "detector_name": "pilatus",
        "transmission": 1.0,
        "num_images": 10,
        "pump_status": True,
        "laser_dwell_s": 0.01,
        "laser_delay_s": 0.005,
    }
    return ExtruderParameters(**params_pp)


@pytest.fixture
def dummy_beam_settings():
    return BeamSettings(
        wavelength_in_a=0.6, beam_size_in_um=(7, 7), beam_center_in_mm=(120.4, 127.6)
    )


def fake_generator(value):
    yield from bps.null()
    return value


@patch(
    "mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.get_detector_type"
)
@patch("mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.caget")
@patch(
    "mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2._read_visit_directory_from_file"
)
@patch(
    "mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.SSX_LOGGER"
)
@patch("mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.bps.rd")
def test_read_parameters(
    fake_rd,
    fake_log,
    mock_read_visit,
    fake_caget,
    fake_det,
    detector_stage,
    RE,
):
    mock_attenuator = MagicMock()
    fake_det.side_effect = [fake_generator(Eiger())]
    fake_rd.side_effect = [fake_generator(0.3)]
    with patch(
        "mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.ExtruderParameters",
    ):
        RE(read_parameters(detector_stage, mock_attenuator))

    assert fake_caget.call_count == 8
    fake_log.warning.assert_called_once()
    assert fake_log.info.call_count == 3


@patch("mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.caget")
@patch("mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.caput")
@patch(
    "mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.get_detector_type"
)
@patch(
    "mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.SSX_LOGGER"
)
def test_initialise_extruder(
    fake_log,
    fake_det,
    fake_caput,
    fake_caget,
    detector_stage,
    RE,
):
    fake_caget.return_value = "/path/to/visit"
    fake_det.side_effect = [fake_generator(Eiger())]
    RE(initialise_extruder(detector_stage))
    assert fake_caput.call_count == 10
    assert fake_caget.call_count == 1


async def test_enterhutch(detector_stage, RE):
    RE(enter_hutch(detector_stage))
    assert await detector_stage.z.user_setpoint.get_value() == 1480


@pytest.mark.parametrize(
    "laser_mode, det_type, expected_in1, expected_out",
    [
        ("laseron", Eiger(), "Yes", I24_ZEBRA_MAPPING.sources.SOFT_IN3),
        ("laseroff", Eiger(), "No", I24_ZEBRA_MAPPING.sources.DISCONNECT),
        ("laseron", Pilatus(), "Yes", I24_ZEBRA_MAPPING.sources.SOFT_IN3),
        ("laseroff", Pilatus(), "No", I24_ZEBRA_MAPPING.sources.DISCONNECT),
    ],
)
@patch(
    "mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.get_detector_type"
)
async def test_laser_check(
    fake_det,
    laser_mode,
    expected_in1,
    expected_out,
    det_type,
    zebra,
    detector_stage,
    RE,
):
    fake_det.side_effect = [fake_generator(det_type)]
    RE(laser_check(laser_mode, zebra, detector_stage))

    TTL = (
        I24_ZEBRA_MAPPING.outputs.TTL_EIGER
        if isinstance(det_type, Pilatus)
        else I24_ZEBRA_MAPPING.outputs.TTL_PILATUS
    )
    assert await zebra.inputs.soft_in_1.get_value() == expected_in1
    assert await zebra.output.out_pvs[TTL].get_value() == expected_out


@patch(
    "mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.bps.sleep"
)
@patch("mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.DCID")
@patch(
    "mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.call_nexgen"
)
@patch("mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.caput")
@patch("mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.caget")
@patch("mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.sup")
@patch(
    "mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.setup_zebra_for_quickshot_plan"
)
@patch("mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.bps.rd")
@patch(
    "mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.read_beam_info_from_hardware"
)
def test_run_extruder_quickshot_with_eiger(
    mock_read_beam_info,
    fake_read,
    mock_quickshot_plan,
    fake_sup,
    fake_caget,
    fake_caput,
    fake_nexgen,
    fake_dcid,
    fake_sleep,
    RE,
    zebra,
    shutter,
    aperture,
    backlight,
    beamstop,
    detector_stage,
    dcm,
    mirrors,
    eiger_beam_center,
    dummy_params,
    dummy_beam_settings,
):
    fake_start_time = MagicMock()
    mock_read_beam_info.side_effect = [fake_generator(dummy_beam_settings)]
    # Mock end of data collection (zebra disarmed)
    fake_read.side_effect = [
        fake_generator(1605),  # beam center
        fake_generator(1702),
        fake_generator(0),  # zebra disarm
    ]
    with patch(
        "mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.BEAM_CENTER_LUT_FILES",
        new=TEST_LUT,
    ):
        RE(
            main_extruder_plan(
                zebra,
                aperture,
                backlight,
                beamstop,
                detector_stage,
                shutter,
                dcm,
                mirrors,
                eiger_beam_center,
                dummy_params,
                fake_dcid,
                fake_start_time,
            )
        )
    fake_nexgen.assert_called_once_with(
        None, dummy_params, 0.6, (1605, 1702), fake_start_time
    )
    assert fake_dcid.generate_dcid.call_count == 1
    assert fake_dcid.notify_start.call_count == 1
    assert fake_sup.setup_beamline_for_collection_plan.call_count == 1
    mock_quickshot_plan.assert_called_once()
    mock_read_beam_info.assert_called_once()


@patch(
    "mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.bps.sleep"
)
@patch("mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.DCID")
@patch("mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.caput")
@patch("mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.caget")
@patch("mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.sup")
@patch(
    "mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.setup_zebra_for_extruder_with_pump_probe_plan"
)
@patch("mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.bps.rd")
@patch(
    "mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.read_beam_info_from_hardware"
)
@patch(
    "mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.get_pilatus_filename_template_from_device"
)
def test_run_extruder_pump_probe_with_pilatus(
    mock_pilatus_temp,
    mock_read_beam_info,
    fake_read,
    mock_pp_plan,
    fake_sup,
    fake_caget,
    fake_caput,
    fake_dcid,
    fake_sleep,
    RE,
    zebra,
    shutter,
    aperture,
    backlight,
    beamstop,
    detector_stage,
    dcm,
    mirrors,
    pilatus_beam_center,
    dummy_params_pp,
):
    fake_start_time = MagicMock()
    set_mock_value(dcm.wavelength_in_a, 0.6)
    # Mock end of data collection (zebra disarmed)
    fake_read.side_effect = [fake_generator(0)]
    mock_pilatus_temp.side_effect = [fake_generator("test_00001_#####.cbf")]
    with patch(
        "mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.BEAM_CENTER_LUT_FILES",
        new=TEST_LUT,
    ):
        RE(
            main_extruder_plan(
                zebra,
                aperture,
                backlight,
                beamstop,
                detector_stage,
                shutter,
                dcm,
                mirrors,
                pilatus_beam_center,
                dummy_params_pp,
                fake_dcid,
                fake_start_time,
            )
        )
    mock_pilatus_temp.assert_called_once()
    assert fake_dcid.generate_dcid.call_count == 1
    assert fake_dcid.notify_start.call_count == 1
    assert fake_sup.move_detector_stage_to_position_plan.call_count == 1
    mock_pp_plan.assert_called_once()

    shutter_call_list = [
        call("Reset", wait=True),
        call("Open", wait=True),
    ]
    mock_shutter = get_mock_put(shutter.control)
    mock_shutter.assert_has_calls(shutter_call_list)
    mock_read_beam_info.assert_called_once()


@patch(
    "mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.reset_zebra_when_collection_done_plan"
)
@patch("mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.DCID")
@patch("mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.caput")
@patch("mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.caget")
@patch("mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.sup")
def test_tidy_up_at_collection_end_plan_with_eiger(
    fake_sup,
    fake_caget,
    fake_caput,
    fake_dcid,
    mock_reset_zebra_plan,
    RE,
    zebra,
    shutter,
    dummy_params,
):
    RE(tidy_up_at_collection_end_plan(zebra, shutter, dummy_params, fake_dcid))

    mock_reset_zebra_plan.assert_called_once()
    mock_shutter = get_mock_put(shutter.control)
    mock_shutter.assert_has_calls([call("Close", wait=True)])

    assert fake_dcid.notify_end.call_count == 1
    assert fake_caget.call_count == 1

    fake_sup.eiger.assert_called_once_with("return-to-normal", None)


@patch(
    "mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.bps.sleep"
)
@patch("mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.caput")
@patch("mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.DCID")
@patch(
    "mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.disarm_zebra"
)
def test_aborted_plan_with_pilatus(
    mock_disarm, fake_dcid, fake_caput, fake_sleep, RE, zebra
):
    RE(collection_aborted_plan(zebra, "pilatus", fake_dcid))

    mock_disarm.assert_called_once()
    fake_caput.assert_has_calls([call(ANY, 0)])
    fake_dcid.collection_complete.assert_called_once_with(ANY, aborted=True)


@patch("mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.DCID")
@patch(
    "mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.bps.sleep"
)
@patch("mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.caput")
def test_collection_complete_plan_with_eiger(
    fake_caput, fake_sleep, fake_dcid, dummy_params, RE
):
    RE(
        collection_complete_plan(
            dummy_params.collection_directory, dummy_params.detector_name, fake_dcid
        )
    )

    call_list = [call(ANY, 0), call(ANY, "Done")]
    fake_caput.assert_has_calls(call_list)

    fake_dcid.collection_complete.assert_called_once_with(ANY, aborted=False)



from __future__ import annotations

import dataclasses

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
import pydantic
from blueapi.core import BlueskyContext
from bluesky.utils import MsgGenerator
from dodal.devices.aperturescatterguard import ApertureScatterguard
from dodal.devices.attenuator.attenuator import BinaryFilterAttenuator
from dodal.devices.backlight import Backlight
from dodal.devices.dcm import DCM
from dodal.devices.detector.detector_motion import DetectorMotion
from dodal.devices.eiger import EigerDetector
from dodal.devices.flux import Flux
from dodal.devices.i03.beamstop import Beamstop
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.oav.oav_parameters import OAVParameters
from dodal.devices.robot import BartRobot
from dodal.devices.s4_slit_gaps import S4SlitGaps
from dodal.devices.smargon import Smargon
from dodal.devices.synchrotron import Synchrotron
from dodal.devices.undulator import Undulator
from dodal.devices.xbpm_feedback import XBPMFeedback
from dodal.devices.zebra.zebra import RotationDirection, Zebra
from dodal.devices.zebra.zebra_controlled_shutter import ZebraShutter
from dodal.plan_stubs.check_topup import check_topup_and_wait_if_necessary
from dodal.plans.preprocessors.verify_undulator_gap import (
    verify_undulator_gap_before_run_decorator,
)

from mx_bluesky.common.parameters.components import WithSnapshot
from mx_bluesky.common.plans.read_hardware import (
    read_hardware_for_zocalo,
    standard_read_hardware_during_collection,
    standard_read_hardware_pre_collection,
)
from mx_bluesky.common.preprocessors.preprocessors import (
    transmission_and_xbpm_feedback_for_collection_decorator,
)
from mx_bluesky.common.utils.context import device_composite_from_context
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.hyperion.device_setup_plans.manipulate_sample import (
    cleanup_sample_environment,
    move_phi_chi_omega,
    move_x_y_z,
    setup_sample_environment,
)
from mx_bluesky.hyperion.device_setup_plans.setup_zebra import (
    arm_zebra,
    setup_zebra_for_rotation,
    tidy_up_zebra_after_rotation_scan,
)
from mx_bluesky.hyperion.device_setup_plans.utils import (
    start_preparing_data_collection_then_do_plan,
)
from mx_bluesky.hyperion.experiment_plans.oav_snapshot_plan import (
    OavSnapshotComposite,
    oav_snapshot_plan,
    setup_beamline_for_OAV,
)
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.rotation import (
    MultiRotationScan,
    RotationScan,
)


@pydantic.dataclasses.dataclass(config={"arbitrary_types_allowed": True})
class RotationScanComposite(OavSnapshotComposite):
    """All devices which are directly or indirectly required by this plan"""

    aperture_scatterguard: ApertureScatterguard
    attenuator: BinaryFilterAttenuator
    backlight: Backlight
    beamstop: Beamstop
    dcm: DCM
    detector_motion: DetectorMotion
    eiger: EigerDetector
    flux: Flux
    robot: BartRobot
    smargon: Smargon
    undulator: Undulator
    synchrotron: Synchrotron
    s4_slit_gaps: S4SlitGaps
    sample_shutter: ZebraShutter
    zebra: Zebra
    oav: OAV
    xbpm_feedback: XBPMFeedback


def create_devices(context: BlueskyContext) -> RotationScanComposite:
    """Ensures necessary devices have been instantiated"""

    return device_composite_from_context(context, RotationScanComposite)


DEFAULT_DIRECTION = RotationDirection.NEGATIVE
DEFAULT_MAX_VELOCITY = 120
# Use a slightly larger time to acceleration than EPICS as it's better to be cautious
ACCELERATION_MARGIN = 1.5


@dataclasses.dataclass
class RotationMotionProfile:
    start_scan_deg: float
    start_motion_deg: float
    scan_width_deg: float
    shutter_time_s: float
    direction: RotationDirection
    speed_for_rotation_deg_s: float
    acceleration_offset_deg: float
    shutter_opening_deg: float
    total_exposure_s: float
    distance_to_move_deg: float
    max_velocity_deg_s: float


def calculate_motion_profile(
    params: RotationScan,
    motor_time_to_speed_s: float,
    max_velocity_deg_s: float,
) -> RotationMotionProfile:
    """Calculates the various numbers needed for motions in the rotation scan.
    Rotates through "scan width" plus twice an "offset" to take into account
    acceleration at the start and deceleration at the end, plus the number of extra
    degrees of rotation needed to make sure the fast shutter has fully opened before the
    detector trigger is sent.
    See https://github.com/DiamondLightSource/hyperion/wiki/rotation-scan-geometry
    for a simple pictorial explanation."""

    assert params.rotation_increment_deg > 0

    direction = params.rotation_direction
    start_scan_deg = params.omega_start_deg

    if params.features.omega_flip:
        # If omega_flip is True then the motor omega axis is inverted with respect to the
        # hyperion coordinate system.
        start_scan_deg = -start_scan_deg
        direction = (
            direction.POSITIVE
            if direction == direction.NEGATIVE
            else direction.NEGATIVE
        )

    num_images = params.num_images
    shutter_time_s = params.shutter_opening_time_s
    image_width_deg = params.rotation_increment_deg
    exposure_time_s = params.exposure_time_s
    motor_time_to_speed_s *= ACCELERATION_MARGIN

    LOGGER.info("Calculating rotation scan motion profile:")
    LOGGER.info(
        f"{num_images=}, {shutter_time_s=}, {image_width_deg=}, {exposure_time_s=}, {direction=}"
    )

    scan_width_deg = num_images * params.rotation_increment_deg
    LOGGER.info(f"{scan_width_deg=} = {num_images=} * {params.rotation_increment_deg=}")

    speed_for_rotation_deg_s = image_width_deg / exposure_time_s
    LOGGER.info("speed_for_rotation_deg_s = image_width_deg / exposure_time_s")
    LOGGER.info(
        f"{speed_for_rotation_deg_s=} = {image_width_deg=} / {exposure_time_s=}"
    )

    acceleration_offset_deg = motor_time_to_speed_s * speed_for_rotation_deg_s
    LOGGER.info(
        f"{acceleration_offset_deg=} = {motor_time_to_speed_s=} * {speed_for_rotation_deg_s=}"
    )

    start_motion_deg = start_scan_deg - (acceleration_offset_deg * direction.multiplier)
    LOGGER.info(
        f"{start_motion_deg=} = {start_scan_deg=} - ({acceleration_offset_deg=} * {direction.multiplier=})"
    )

    shutter_opening_deg = speed_for_rotation_deg_s * shutter_time_s
    LOGGER.info(
        f"{shutter_opening_deg=} = {speed_for_rotation_deg_s=} * {shutter_time_s=}"
    )

    shutter_opening_deg = speed_for_rotation_deg_s * shutter_time_s
    LOGGER.info(
        f"{shutter_opening_deg=} = {speed_for_rotation_deg_s=} * {shutter_time_s=}"
    )

    total_exposure_s = num_images * exposure_time_s
    LOGGER.info(f"{total_exposure_s=} = {num_images=} * {exposure_time_s=}")

    distance_to_move_deg = (
        scan_width_deg + shutter_opening_deg + acceleration_offset_deg * 2
    ) * direction.multiplier
    LOGGER.info(
        f"{distance_to_move_deg=} = ({scan_width_deg=} + {shutter_opening_deg=} + {acceleration_offset_deg=} * 2) * {direction=})"
    )

    return RotationMotionProfile(
        start_scan_deg=start_scan_deg,
        start_motion_deg=start_motion_deg,
        scan_width_deg=scan_width_deg,
        shutter_time_s=shutter_time_s,
        direction=direction,
        speed_for_rotation_deg_s=speed_for_rotation_deg_s,
        acceleration_offset_deg=acceleration_offset_deg,
        shutter_opening_deg=shutter_opening_deg,
        total_exposure_s=total_exposure_s,
        distance_to_move_deg=distance_to_move_deg,
        max_velocity_deg_s=max_velocity_deg_s,
    )


def rotation_scan_plan(
    composite: RotationScanComposite,
    params: RotationScan,
    motion_values: RotationMotionProfile,
):
    """A stub plan to collect diffraction images from a sample continuously rotating
    about a fixed axis - for now this axis is limited to omega.
    Needs additional setup of the sample environment and a wrapper to clean up."""

    @bpp.set_run_key_decorator(CONST.PLAN.ROTATION_MAIN)
    @bpp.run_decorator(
        md={
            "subplan_name": CONST.PLAN.ROTATION_MAIN,
            "scan_points": [params.scan_points],
        }
    )
    def _rotation_scan_plan(
        motion_values: RotationMotionProfile,
        composite: RotationScanComposite,
    ):
        axis = composite.smargon.omega

        # can move to start as fast as possible
        yield from bps.abs_set(
            axis.velocity, motion_values.max_velocity_deg_s, wait=True
        )
        LOGGER.info(f"moving omega to beginning, {motion_values.start_scan_deg=}")
        yield from bps.abs_set(
            axis,
            motion_values.start_motion_deg,
            group=CONST.WAIT.ROTATION_READY_FOR_DC,
        )

        yield from setup_zebra_for_rotation(
            composite.zebra,
            composite.sample_shutter,
            start_angle=motion_values.start_scan_deg,
            scan_width=motion_values.scan_width_deg,
            direction=motion_values.direction,
            shutter_opening_deg=motion_values.shutter_opening_deg,
            shutter_opening_s=motion_values.shutter_time_s,
            group="setup_zebra",
        )

        yield from setup_sample_environment(
            composite.aperture_scatterguard,
            params.selected_aperture,
            composite.backlight,
            group=CONST.WAIT.ROTATION_READY_FOR_DC,
        )

        LOGGER.info("Wait for any previous moves...")
        # wait for all the setup tasks at once
        yield from bps.wait(CONST.WAIT.ROTATION_READY_FOR_DC)
        yield from bps.wait(CONST.WAIT.MOVE_GONIO_TO_START)

        # get some information for the ispyb deposition and trigger the callback
        yield from read_hardware_for_zocalo(composite.eiger)

        yield from standard_read_hardware_pre_collection(
            composite.undulator,
            composite.synchrotron,
            composite.s4_slit_gaps,
            composite.dcm,
            composite.smargon,
        )

        # Get ready for the actual scan
        yield from bps.abs_set(
            axis.velocity, motion_values.speed_for_rotation_deg_s, wait=True
        )

        yield from bps.wait("setup_zebra")
        yield from arm_zebra(composite.zebra)

        # Check topup gate
        yield from check_topup_and_wait_if_necessary(
            composite.synchrotron,
            motion_values.total_exposure_s,
            ops_time=10.0,  # Additional time to account for rotation, is s
        )  # See #https://github.com/DiamondLightSource/hyperion/issues/932

        LOGGER.info("Executing rotation scan")
        yield from bps.rel_set(axis, motion_values.distance_to_move_deg, wait=True)

        yield from standard_read_hardware_during_collection(
            composite.aperture_scatterguard,
            composite.attenuator,
            composite.flux,
            composite.dcm,
            composite.eiger,
        )

    yield from _rotation_scan_plan(motion_values, composite)


def _cleanup_plan(composite: RotationScanComposite, **kwargs):
    LOGGER.info("Cleaning up after rotation scan")
    max_vel = yield from bps.rd(composite.smargon.omega.max_velocity)
    yield from cleanup_sample_environment(composite.detector_motion, group="cleanup")
    yield from bps.abs_set(composite.smargon.omega.velocity, max_vel, group="cleanup")
    yield from tidy_up_zebra_after_rotation_scan(
        composite.zebra, composite.sample_shutter, group="cleanup", wait=False
    )
    yield from bps.wait("cleanup")


def _move_and_rotation(
    composite: RotationScanComposite,
    params: RotationScan,
    oav_params: OAVParameters,
):
    motor_time_to_speed = yield from bps.rd(composite.smargon.omega.acceleration_time)
    max_vel = yield from bps.rd(composite.smargon.omega.max_velocity)
    motion_values = calculate_motion_profile(params, motor_time_to_speed, max_vel)

    def _div_by_1000_if_not_none(num: float | None):
        return num / 1000 if num else num

    LOGGER.info("moving to position (if specified)")
    yield from move_x_y_z(
        composite.smargon,
        _div_by_1000_if_not_none(params.x_start_um),
        _div_by_1000_if_not_none(params.y_start_um),
        _div_by_1000_if_not_none(params.z_start_um),
        group=CONST.WAIT.MOVE_GONIO_TO_START,
    )
    yield from move_phi_chi_omega(
        composite.smargon,
        params.phi_start_deg,
        params.chi_start_deg,
        group=CONST.WAIT.MOVE_GONIO_TO_START,
    )
    if params.take_snapshots:
        yield from bps.wait(CONST.WAIT.MOVE_GONIO_TO_START)
        yield from setup_beamline_for_OAV(
            composite.smargon, composite.backlight, composite.aperture_scatterguard
        )
        yield from bps.wait(group=CONST.WAIT.READY_FOR_OAV)
        if params.selected_aperture:
            yield from bps.prepare(
                composite.aperture_scatterguard,
                params.selected_aperture,
                group=CONST.WAIT.ROTATION_READY_FOR_DC,
            )
        yield from oav_snapshot_plan(composite, params, oav_params)
    yield from rotation_scan_plan(
        composite,
        params,
        motion_values,
    )


def rotation_scan(
    composite: RotationScanComposite,
    parameters: RotationScan,
    oav_params: OAVParameters | None = None,
) -> MsgGenerator:
    parameters.features.update_self_from_server()

    if not oav_params:
        oav_params = OAVParameters(context="xrayCentring")

    @transmission_and_xbpm_feedback_for_collection_decorator(
        composite,
        parameters.transmission_frac,
    )
    @verify_undulator_gap_before_run_decorator(composite)
    @bpp.set_run_key_decorator("rotation_scan")
    @bpp.run_decorator(  # attach experiment metadata to the start document
        md={
            "subplan_name": CONST.PLAN.ROTATION_OUTER,
            "mx_bluesky_parameters": parameters.model_dump_json(),
            "with_snapshot": parameters.model_dump_json(
                include=WithSnapshot.model_fields.keys()  # type: ignore
            ),
            "activate_callbacks": [
                "BeamDrawingCallback",
                "RotationISPyBCallback",
                "RotationNexusFileCallback",
            ],
        }
    )
    def rotation_scan_plan_with_stage_and_cleanup(
        params: RotationScan,
    ):
        eiger: EigerDetector = composite.eiger
        eiger.set_detector_parameters(params.detector_params)

        @bpp.finalize_decorator(lambda: _cleanup_plan(composite))
        def rotation_with_cleanup_and_stage(params: RotationScan):
            yield from _move_and_rotation(composite, params, oav_params)

        LOGGER.info("setting up and staging eiger...")
        yield from start_preparing_data_collection_then_do_plan(
            composite.beamstop,
            eiger,
            composite.detector_motion,
            params.detector_distance_mm,
            rotation_with_cleanup_and_stage(params),
            group=CONST.WAIT.ROTATION_READY_FOR_DC,
        )
        yield from bps.unstage(eiger)

    yield from rotation_scan_plan_with_stage_and_cleanup(parameters)


def multi_rotation_scan(
    composite: RotationScanComposite,
    parameters: MultiRotationScan,
    oav_params: OAVParameters | None = None,
) -> MsgGenerator:
    parameters.features.update_self_from_server()
    if not oav_params:
        oav_params = OAVParameters(context="xrayCentring")
    eiger: EigerDetector = composite.eiger
    eiger.set_detector_parameters(parameters.detector_params)

    @transmission_and_xbpm_feedback_for_collection_decorator(
        composite,
        parameters.transmission_frac,
    )
    @bpp.set_run_key_decorator("multi_rotation_scan")
    @bpp.run_decorator(
        md={
            "subplan_name": CONST.PLAN.ROTATION_MULTI,
            "full_num_of_images": parameters.num_images,
            "meta_data_run_number": parameters.detector_params.run_number,
            "activate_callbacks": [
                "RotationISPyBCallback",
                "RotationNexusFileCallback",
            ],
        }
    )
    @bpp.finalize_decorator(lambda: _cleanup_plan(composite))
    def _multi_rotation_scan():
        for single_scan in parameters.single_rotation_scans:

            @verify_undulator_gap_before_run_decorator(composite)
            @bpp.set_run_key_decorator("rotation_scan")
            @bpp.run_decorator(  # attach experiment metadata to the start document
                md={
                    "subplan_name": CONST.PLAN.ROTATION_OUTER,
                    "mx_bluesky_parameters": single_scan.model_dump_json(),
                }
            )
            def rotation_scan_core(
                params: RotationScan,
            ):
                yield from _move_and_rotation(composite, params, oav_params)

            yield from rotation_scan_core(single_scan)

        yield from bps.unstage(eiger)

    LOGGER.info("setting up and staging eiger...")
    yield from start_preparing_data_collection_then_do_plan(
        composite.beamstop,
        eiger,
        composite.detector_motion,
        parameters.detector_distance_mm,
        _multi_rotation_scan(),
        group=CONST.WAIT.ROTATION_READY_FOR_DC,
    )



import asyncio
from datetime import datetime
from unittest.mock import ANY, MagicMock, call, mock_open, patch

import bluesky.plan_stubs as bps
import pytest
from bluesky.utils import FailedStatus
from dodal.devices.hutch_shutter import HutchShutter
from dodal.devices.i24.pmac import PMAC
from dodal.devices.zebra.zebra import Zebra
from ophyd_async.testing import callback_on_mock_put, get_mock_put, set_mock_value

from mx_bluesky.beamlines.i24.serial.fixed_target.ft_utils import (
    ChipType,
    MappingType,
    PumpProbeSetting,
)
from mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1 import (
    PMAC_MOVE_TIME,
    calculate_collection_timeout,
    finish_i24,
    get_chip_prog_values,
    get_prog_num,
    kickoff_and_complete_collection,
    load_motion_program_data,
    main_fixed_target_plan,
    run_aborted_plan,
    run_fixed_target_plan,
    set_datasize,
    start_i24,
    tidy_up_after_collection_plan,
    write_userlog,
)
from mx_bluesky.beamlines.i24.serial.parameters.experiment_parameters import (
    BeamSettings,
)

from ..conftest import TEST_LUT

chipmap_str = """01status    P3011       1
02status    P3021       0
03status    P3031       0
04status    P3041       0"""


def fake_generator(value):
    yield from bps.null()
    return value


def test_calculate_collection_timeout(dummy_params_without_pp):
    assert dummy_params_without_pp.total_num_images == 400
    expected_collection_time = (
        dummy_params_without_pp.total_num_images
        * dummy_params_without_pp.exposure_time_s
    )
    buffer = dummy_params_without_pp.total_num_images * PMAC_MOVE_TIME + 600
    timeout = calculate_collection_timeout(dummy_params_without_pp)

    assert timeout == expected_collection_time + buffer


def test_calculate_collection_timeout_for_eava(dummy_params_with_pp):
    assert dummy_params_with_pp.total_num_images == 400
    buffer = dummy_params_with_pp.total_num_images * PMAC_MOVE_TIME + 600
    expected_pump_and_probe_time = 12.05
    timeout = calculate_collection_timeout(dummy_params_with_pp)

    assert timeout == expected_pump_and_probe_time + buffer


@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.SSX_LOGGER"
)
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.Path.mkdir"
)
def test_write_userlog(fake_mkdir, fake_log, dummy_params_without_pp):
    with patch(
        "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.open",
        mock_open(),
    ):
        write_userlog(dummy_params_without_pp, "some_file", 1.0, 0.6)
    fake_log.debug.assert_called_once()


@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.SSX_LOGGER"
)
@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.caput")
def test_set_datasize_for_one_block_and_two_exposures(
    fake_caput, fake_log, dummy_params_without_pp
):
    dummy_params_without_pp.num_exposures = 2
    dummy_params_without_pp.chip_map = [20]
    assert dummy_params_without_pp.total_num_images == 800
    set_datasize(dummy_params_without_pp)
    fake_caput.assert_called_once_with("ME14E-MO-IOC-01:GP10", 800)
    assert fake_log.info.call_count == 1
    assert fake_log.debug.call_count == 4


def test_get_chip_prog_values(dummy_params_without_pp):
    dummy_params_without_pp.num_exposures = 2
    chip_dict = get_chip_prog_values(
        dummy_params_without_pp,
    )
    assert isinstance(chip_dict, dict)
    assert chip_dict["X_NUM_STEPS"][1] == 20 and chip_dict["X_NUM_BLOCKS"][1] == 8
    assert chip_dict["PUMP_REPEAT"][1] == 0
    assert chip_dict["N_EXPOSURES"][1] == 2


@pytest.mark.parametrize(
    "chip_type, map_type, pump_repeat, expected_prog",
    [
        (ChipType.Oxford, MappingType.NoMap, PumpProbeSetting.NoPP, 11),
        (ChipType.Oxford, MappingType.Lite, PumpProbeSetting.NoPP, 12),
        (ChipType.OxfordInner, MappingType.Lite, PumpProbeSetting.NoPP, 12),
        (ChipType.Custom, MappingType.Lite, PumpProbeSetting.NoPP, 11),
        (ChipType.Minichip, MappingType.NoMap, PumpProbeSetting.NoPP, 11),
        (ChipType.Oxford, MappingType.Lite, PumpProbeSetting.Short2, 14),
        (ChipType.Minichip, MappingType.NoMap, PumpProbeSetting.Repeat5, 14),
        (ChipType.Custom, MappingType.Lite, PumpProbeSetting.Medium1, 14),
    ],
)
def test_get_prog_number(chip_type, map_type, pump_repeat, expected_prog):
    assert get_prog_num(chip_type, map_type, pump_repeat) == expected_prog


@pytest.mark.parametrize(
    "map_type, pump_repeat, checker, expected_calls",
    [
        (0, 0, False, ["P1100=1"]),  # Full chip, no pump probe, no checker
        (1, 0, False, ["P1200=1"]),  # Mapping lite, no pp, no checker
        (
            1,
            2,
            False,
            ["P1439=0", "P1441=0", "P1400=1"],
        ),  # Map irrelevant, pp to Repeat1, no checker
        (
            0,
            3,
            True,
            ["P1439=0", "P1439=1", "P1441=0", "P1400=1"],
        ),  # Map irrelevant, pp to Repeat2, checker enabled
        (
            1,
            8,
            False,
            ["P1439=0", "P1441=50", "P1400=1"],
        ),  # Map irrelevant, pp to Medium1, checker disabled
    ],
)
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.bps.sleep"
)
def test_load_motion_program_data(
    mock_sleep,
    map_type: int,
    pump_repeat: int,
    checker: bool,
    expected_calls: list,
    pmac: PMAC,
    RE,
):
    test_dict = {"N_EXPOSURES": [0, 1]}
    RE(load_motion_program_data(pmac, test_dict, map_type, pump_repeat, checker))
    call_list = []
    for i in expected_calls:
        call_list.append(call(i, wait=True))
    mock_pmac_str = get_mock_put(pmac.pmac_string)
    mock_pmac_str.assert_has_calls(call_list)


@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.DCID")
@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.caput")
@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.caget")
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.cagetstring"
)
@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.sup")
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.bps.sleep"
)
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.datetime"
)
def test_start_i24_with_eiger(
    fake_datetime,
    fake_sleep,
    fake_sup,
    fake_cagetstring,
    fake_caget,
    fake_caput,
    fake_dcid,
    zebra: Zebra,
    shutter: HutchShutter,
    RE,
    aperture,
    backlight,
    beamstop,
    detector_stage,
    dcm,
    mirrors,
    eiger_beam_center,
    dummy_params_without_pp,
):
    expected_start = datetime.now()
    fake_datetime.now.return_value = expected_start
    dummy_params_without_pp.chip_map = [1, 2]
    assert dummy_params_without_pp.total_num_images == 800
    set_mock_value(dcm.wavelength_in_a, 0.6)
    expected_beam_settings = BeamSettings(
        wavelength_in_a=0.6,
        beam_size_in_um=(7.0, 7.0),
        beam_center_in_mm=(1605 * 0.075, 1702 * 0.075),
    )
    expected_odin_filename = f"{dummy_params_without_pp.filename}_0001"
    fake_cagetstring.return_value = expected_odin_filename

    RE(
        start_i24(
            zebra,
            aperture,
            backlight,
            beamstop,
            detector_stage,
            shutter,
            dummy_params_without_pp,
            dcm,
            mirrors,
            eiger_beam_center,
            fake_dcid,
        )
    )
    assert fake_sup.eiger.call_count == 1
    assert fake_sup.setup_beamline_for_collection_plan.call_count == 1
    assert fake_sup.move_detector_stage_to_position_plan.call_count == 1
    fake_cagetstring.assert_called_once()
    fake_dcid.generate_dcid.assert_called_with(
        beam_settings=expected_beam_settings,
        image_dir=dummy_params_without_pp.collection_directory.as_posix(),
        file_template=f"{expected_odin_filename}.nxs",
        num_images=dummy_params_without_pp.total_num_images,
        shots_per_position=dummy_params_without_pp.num_exposures,
        start_time=expected_start,
        pump_probe=False,
    )

    shutter_call_list = [
        call("Reset", wait=True),
        call("Open", wait=True),
    ]
    mock_shutter = get_mock_put(shutter.control)
    mock_shutter.assert_has_calls(shutter_call_list)


@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.write_userlog"
)
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.bps.sleep"
)
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.cagetstring"
)
@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.caget")
@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.sup")
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.reset_zebra_when_collection_done_plan"
)
@patch("mx_bluesky.beamlines.i24.serial.extruder.i24ssx_Extruder_Collect_py3v2.bps.rd")
def test_finish_i24(
    fake_read,
    fake_reset_zebra,
    fake_sup,
    fake_caget,
    fake_cagetstring,
    fake_sleep,
    fake_userlog,
    zebra,
    pmac,
    shutter,
    dcm,
    dummy_params_without_pp,
    RE,
):
    fake_read.side_effect = [fake_generator(0.6)]
    fake_caget.return_value = 0.0
    fake_cagetstring.return_value = "chip_01"
    RE(finish_i24(zebra, pmac, shutter, dcm, dummy_params_without_pp))

    fake_reset_zebra.assert_called_once()

    fake_sup.eiger.assert_called_once_with("return-to-normal", None)

    mock_pmac_string = get_mock_put(pmac.pmac_string)
    mock_pmac_string.assert_has_calls([call("!x0y0z0", wait=True)])

    mock_shutter = get_mock_put(shutter.control)
    mock_shutter.assert_has_calls([call("Close", wait=True)])

    fake_userlog.assert_called_once_with(dummy_params_without_pp, "chip_01", 0.0, 0.6)


@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.DCID")
def test_run_aborted_plan(fake_dcid: MagicMock, pmac: PMAC, RE, done_status):
    pmac.abort_program.trigger = MagicMock(return_value=done_status)
    RE(run_aborted_plan(pmac, fake_dcid))

    pmac.abort_program.trigger.assert_called_once()
    fake_dcid.collection_complete.assert_called_once_with(ANY, aborted=True)


@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.finish_i24"
)
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.bps.sleep"
)
@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.DCID")
@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.caput")
async def test_tidy_up_after_collection_plan(
    fake_caput,
    fake_dcid,
    fake_sleep,
    mock_finish,
    zebra,
    pmac,
    shutter,
    dcm,
    RE,
    dummy_params_without_pp,
):
    RE(
        tidy_up_after_collection_plan(
            zebra, pmac, shutter, dcm, dummy_params_without_pp, fake_dcid
        )
    )
    assert await zebra.inputs.soft_in_2.get_value() == "No"

    fake_dcid.notify_end.assert_called_once()

    fake_caput.assert_has_calls([call(ANY, 0), call(ANY, "Done")])

    mock_finish.assert_called_once()


@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.calculate_collection_timeout"
)
async def test_kick_off_and_complete_collection(
    fake_collection_time, pmac, dummy_params_with_pp, RE, done_status
):
    pmac.run_program.kickoff = MagicMock(return_value=done_status)
    pmac.run_program.complete = MagicMock(return_value=done_status)

    async def go_high_then_low():
        set_mock_value(pmac.scanstatus, 1)
        await asyncio.sleep(0.1)
        set_mock_value(pmac.scanstatus, 0)

    callback_on_mock_put(
        pmac.pmac_string,
        lambda *args, **kwargs: asyncio.create_task(go_high_then_low()),  # type: ignore
    )
    fake_collection_time.return_value = 2.0
    res = RE(kickoff_and_complete_collection(pmac, dummy_params_with_pp))

    assert await pmac.program_number.get_value() == 14
    assert await pmac.collection_time.get_value() == 2.0

    pmac.run_program.kickoff.assert_called_once()
    pmac.run_program.complete.assert_called_once()

    assert res.exit_status == "success"


@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.calculate_collection_timeout"
)
@patch("dodal.devices.i24.pmac.DEFAULT_TIMEOUT", 0.1)
async def test_kickoff_and_complete_fails_if_scan_status_pv_does_not_change(
    fake_collection_time, pmac, dummy_params_without_pp, RE
):
    fake_collection_time.return_value = 1.0
    pmac.run_program.KICKOFF_TIMEOUT = 0.1
    set_mock_value(pmac.scanstatus, 0)
    with pytest.raises(FailedStatus):
        RE(kickoff_and_complete_collection(pmac, dummy_params_without_pp))


@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.kickoff_and_complete_collection"
)
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.start_i24"
)
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.set_datasize"
)
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.load_motion_program_data"
)
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.get_chip_prog_values"
)
@patch("mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.DCID")
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.call_nexgen"
)
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.bps.sleep"
)
async def test_main_fixed_target_plan(
    fake_sleep,
    fake_nexgen,
    fake_dcid,
    mock_get_chip_prog,
    mock_motion_program,
    fake_datasize,
    mock_start,
    mock_kickoff,
    RE,
    zebra,
    pmac,
    aperture,
    backlight,
    beamstop,
    detector_stage,
    shutter,
    dcm,
    mirrors,
    eiger_beam_center,
    dummy_params_without_pp,
):
    mock_get_chip_prog.return_value = MagicMock()
    set_mock_value(dcm.wavelength_in_a, 0.6)
    fake_datasize.return_value = 400
    with patch(
        "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.BEAM_CENTER_LUT_FILES",
        new=TEST_LUT,
    ):
        RE(
            main_fixed_target_plan(
                zebra,
                pmac,
                aperture,
                backlight,
                beamstop,
                detector_stage,
                shutter,
                dcm,
                mirrors,
                eiger_beam_center,
                dummy_params_without_pp,
                fake_dcid,
            )
        )

    mock_beam_x = get_mock_put(eiger_beam_center.beam_x)
    mock_pmac_str = get_mock_put(pmac.pmac_string)
    mock_zebra_input = get_mock_put(zebra.inputs.soft_in_2)

    mock_beam_x.assert_called_once_with(
        pytest.approx(1597.06, 1e-2), wait=True
    )  # Check beam center set
    assert dummy_params_without_pp.total_num_images == 400
    mock_get_chip_prog.assert_called_once_with(dummy_params_without_pp)
    mock_motion_program.asset_called_once()
    mock_start.assert_called_once()
    mock_pmac_str.assert_called_once_with(
        "!x0y0z0", wait=True
    )  # Check pmac moved to start
    assert fake_dcid.notify_start.call_count == 1
    mock_zebra_input.assert_called_once_with(
        "Yes", wait=True
    )  # Check fast shutter open
    fake_nexgen.assert_called_once_with(
        mock_get_chip_prog.return_value,
        dummy_params_without_pp,
        0.6,
        (ANY, ANY),
        None,
    )
    mock_kickoff.assert_called_once_with(
        pmac, dummy_params_without_pp
    )  # Check collection kick off


@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.sup.get_beam_center_device"
)
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.Path.mkdir"
)
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.read_parameters"
)
@patch(
    "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.bps.sleep"
)
def test_setup_tasks_in_run_fixed_target_plan(
    fake_sleep,
    fake_read,
    fake_mkdir,
    mock_beam_center,
    zebra,
    pmac,
    aperture,
    backlight,
    beamstop,
    detector_stage,
    shutter,
    dcm,
    mirrors,
    RE,
    dummy_params_without_pp,
):
    mock_attenuator = MagicMock()
    fake_read.side_effect = [fake_generator(dummy_params_without_pp)]
    with (
        patch(
            "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.bpp.contingency_wrapper"
        ),
        patch(
            "mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.upload_chip_map_to_geobrick"
        ) as patch_upload,
    ):
        RE(
            run_fixed_target_plan(
                zebra,
                pmac,
                aperture,
                backlight,
                beamstop,
                detector_stage,
                shutter,
                dcm,
                mirrors,
                mock_attenuator,
            )
        )
        fake_mkdir.assert_called_once()
        mock_beam_center.assert_called_once()
        patch_upload.assert_called_once_with(pmac, dummy_params_without_pp.chip_map)



import bluesky.plan_stubs as bps
from dodal.devices.i03.beamstop import Beamstop, BeamstopPositions

from mx_bluesky.common.utils.log import LOGGER


class BeamstopException(Exception):
    pass


def check_beamstop(beamstop: Beamstop):
    """
    Check the current position of the beamstop to ensure it is in position for data collection
    Args:
        beamstop: The beamstop device

    Raises:
        BeamstopException: If the beamstop is in any other position than DATA_COLLECTION
    """
    current_pos = yield from bps.rd(beamstop.selected_pos)
    if current_pos != BeamstopPositions.DATA_COLLECTION:
        LOGGER.info(f"Beamstop check failed: position {current_pos}")
        raise BeamstopException(
            f"Beamstop is not DATA_COLLECTION, current state is {current_pos}"
        )

    LOGGER.info("Beamstop check ok")



from unittest.mock import MagicMock, patch

import bluesky.preprocessors as bpp
import pytest
from bluesky.preprocessors import run_decorator
from bluesky.run_engine import RunEngine

from mx_bluesky.common.external_interaction.callbacks.sample_handling.sample_handling_callback import (
    SampleHandlingCallback,
)
from mx_bluesky.common.external_interaction.ispyb.exp_eye_store import BLSampleStatus
from mx_bluesky.common.utils.exceptions import SampleException
from mx_bluesky.hyperion.experiment_plans.flyscan_xray_centre_plan import (
    CrystalNotFoundException,
)

TEST_SAMPLE_ID = 123456


@run_decorator(
    md={
        "metadata": {"sample_id": TEST_SAMPLE_ID},
        "activate_callbacks": ["SampleHandlingCallback"],
    }
)
def plan_with_general_exception(exception_type: type, msg: str):
    yield from []
    raise exception_type(msg)


def plan_for_sample_id(sample_id):
    def plan_with_exception():
        yield from []
        raise SampleException(f"Test exception for sample_id {sample_id}")

    yield from bpp.run_wrapper(
        plan_with_exception(),
        md={
            "metadata": {"sample_id": sample_id},
            "activate_callbacks": ["SampleHandlingCallback"],
        },
    )


def plan_with_exception_from_inner_plan():
    @run_decorator(
        md={
            "metadata": {"sample_id": TEST_SAMPLE_ID},
        }
    )
    def inner_plan():
        yield from []
        raise SampleException("Exception from inner plan")

    @run_decorator(
        md={
            "metadata": {"sample_id": TEST_SAMPLE_ID},
            "activate_callbacks": ["SampleHandlingCallback"],
        }
    )
    @bpp.set_run_key_decorator("outer_plan")
    def outer_plan():
        yield from inner_plan()

    yield from outer_plan()


def plan_with_rethrown_exception():
    @run_decorator(
        md={
            "metadata": {"sample_id": TEST_SAMPLE_ID},
        }
    )
    def inner_plan():
        yield from []
        raise AssertionError("Exception from inner plan")

    @run_decorator(
        md={
            "metadata": {"sample_id": TEST_SAMPLE_ID},
            "activate_callbacks": ["SampleHandlingCallback"],
        }
    )
    @bpp.set_run_key_decorator("outer_plan")
    def outer_plan():
        try:
            yield from inner_plan()
        except AssertionError as e:
            raise SampleException("Exception from outer plan") from e

    yield from outer_plan()


@run_decorator(
    md={
        "metadata": {"sample_id": TEST_SAMPLE_ID},
        "activate_callbacks": ["SampleHandlingCallback"],
    }
)
def plan_with_normal_completion():
    yield from []


@pytest.mark.parametrize(
    "exception_type, expected_sample_status, message",
    [
        [AssertionError, BLSampleStatus.ERROR_BEAMLINE, "Test failure"],
        [SampleException, BLSampleStatus.ERROR_SAMPLE, "Test failure"],
        [CrystalNotFoundException, BLSampleStatus.ERROR_SAMPLE, "Test failure"],
        [AssertionError, BLSampleStatus.ERROR_BEAMLINE, None],
    ],
)
def test_sample_handling_callback_intercepts_general_exception(
    RE: RunEngine,
    exception_type: type,
    expected_sample_status: BLSampleStatus,
    message: str,
):
    callback = SampleHandlingCallback()
    RE.subscribe(callback)

    mock_expeye = MagicMock()
    with (
        patch(
            "mx_bluesky.common.external_interaction.callbacks.sample_handling.sample_handling_callback"
            ".ExpeyeInteraction",
            return_value=mock_expeye,
        ),
        pytest.raises(exception_type),
    ):
        RE(plan_with_general_exception(exception_type, message))
    mock_expeye.update_sample_status.assert_called_once_with(
        TEST_SAMPLE_ID, expected_sample_status
    )


def test_sample_handling_callback_closes_run_normally(RE: RunEngine):
    callback = SampleHandlingCallback()
    RE.subscribe(callback)
    mock_expeye = MagicMock()
    with (
        patch.object(callback, "_record_exception") as record_exception,
        patch(
            "mx_bluesky.common.external_interaction.callbacks.sample_handling.sample_handling_callback"
            ".ExpeyeInteraction",
            return_value=mock_expeye,
        ),
    ):
        RE(plan_with_normal_completion())

    record_exception.assert_not_called()


@patch(
    "mx_bluesky.common.external_interaction.callbacks.sample_handling.sample_handling_callback"
    ".ExpeyeInteraction",
)
def test_sample_handling_callback_resets_sample_id(
    mock_expeye_cls: MagicMock, RE: RunEngine
):
    mock_expeye = mock_expeye_cls.return_value
    callback = SampleHandlingCallback()
    RE.subscribe(callback)

    with pytest.raises(SampleException):
        RE(plan_for_sample_id(TEST_SAMPLE_ID))
    mock_expeye.update_sample_status.assert_called_once_with(
        TEST_SAMPLE_ID, BLSampleStatus.ERROR_SAMPLE
    )
    mock_expeye.reset_mock()

    with pytest.raises(SampleException):
        RE(plan_for_sample_id(TEST_SAMPLE_ID + 1))
    mock_expeye.update_sample_status.assert_called_once_with(
        TEST_SAMPLE_ID + 1, BLSampleStatus.ERROR_SAMPLE
    )


@patch(
    "mx_bluesky.common.external_interaction.callbacks.sample_handling.sample_handling_callback"
    ".ExpeyeInteraction",
)
def test_sample_handling_callback_triggered_only_by_outermost_plan_when_exception_thrown_in_inner_plan(
    mock_expeye_cls: MagicMock, RE: RunEngine
):
    mock_expeye = mock_expeye_cls.return_value
    callback = SampleHandlingCallback()
    RE.subscribe(callback)

    with pytest.raises(SampleException):
        RE(plan_with_exception_from_inner_plan())
    mock_expeye.update_sample_status.assert_called_once_with(
        TEST_SAMPLE_ID, BLSampleStatus.ERROR_SAMPLE
    )


@patch(
    "mx_bluesky.common.external_interaction.callbacks.sample_handling.sample_handling_callback"
    ".ExpeyeInteraction",
)
def test_sample_handling_callback_triggered_only_by_outermost_plan_when_exception_rethrown_from_outermost_plan(
    mock_expeye_cls: MagicMock, RE: RunEngine
):
    mock_expeye = mock_expeye_cls.return_value
    callback = SampleHandlingCallback()
    RE.subscribe(callback)

    with pytest.raises(SampleException):
        RE(plan_with_rethrown_exception())
    mock_expeye.update_sample_status.assert_called_once_with(
        TEST_SAMPLE_ID, BLSampleStatus.ERROR_SAMPLE
    )



from __future__ import annotations

import bluesky.plan_stubs as bps
from bluesky.protocols import Readable
from dodal.devices.aperturescatterguard import ApertureScatterguard
from dodal.devices.attenuator.attenuator import BinaryFilterAttenuator
from dodal.devices.dcm import DCM
from dodal.devices.eiger import EigerDetector
from dodal.devices.flux import Flux
from dodal.devices.s4_slit_gaps import S4SlitGaps
from dodal.devices.smargon import Smargon
from dodal.devices.synchrotron import Synchrotron
from dodal.devices.undulator import Undulator

from mx_bluesky.common.parameters.constants import (
    DocDescriptorNames,
)
from mx_bluesky.common.utils.log import LOGGER


def read_hardware_plan(
    signals: list[Readable],
    event_name: str,
):
    LOGGER.info(f"Reading status of beamline for event, {event_name}")
    yield from bps.create(name=event_name)
    for signal in signals:
        yield from bps.read(signal)
    yield from bps.save()


def read_hardware_for_zocalo(detector: EigerDetector):
    """ "
    If the RunEngine is subscribed to the ZocaloCallback, this plan will also trigger zocalo.
    """
    yield from read_hardware_plan(
        [detector.odin.file_writer.id],  # type: ignore
        DocDescriptorNames.ZOCALO_HW_READ,
    )


def standard_read_hardware_pre_collection(
    undulator: Undulator,
    synchrotron: Synchrotron,
    s4_slit_gaps: S4SlitGaps,
    dcm: DCM,
    smargon: Smargon,
):
    LOGGER.info("Reading status of beamline for callbacks, pre collection.")
    signals_to_read_pre_flyscan = [
        undulator.current_gap,
        synchrotron.synchrotron_mode,
        s4_slit_gaps,
        smargon,
        dcm.energy_in_kev,
    ]
    yield from read_hardware_plan(
        signals_to_read_pre_flyscan, DocDescriptorNames.HARDWARE_READ_PRE
    )


def standard_read_hardware_during_collection(
    aperture_scatterguard: ApertureScatterguard,
    attenuator: BinaryFilterAttenuator,
    flux: Flux,
    dcm: DCM,
    detector: EigerDetector,
):
    signals_to_read_during_collection = [
        aperture_scatterguard,
        attenuator.actual_transmission,
        flux.flux_reading,
        dcm.energy_in_kev,
        detector.bit_depth,
    ]
    yield from read_hardware_plan(
        signals_to_read_during_collection, DocDescriptorNames.HARDWARE_READ_DURING
    )



from event_model import RunStart, RunStop

from mx_bluesky.common.external_interaction.callbacks.common.plan_reactive_callback import (
    PlanReactiveCallback,
)
from mx_bluesky.common.external_interaction.ispyb.exp_eye_store import (
    BLSampleStatus,
    ExpeyeInteraction,
)
from mx_bluesky.common.utils.exceptions import CrystalNotFoundException, SampleException
from mx_bluesky.common.utils.log import ISPYB_ZOCALO_CALLBACK_LOGGER


class SampleHandlingCallback(PlanReactiveCallback):
    """Intercepts exceptions from experiment plans and updates the ISPyB BLSampleStatus
    field according to the type of exception raised."""

    def __init__(self, record_loaded_on_success=False):
        super().__init__(log=ISPYB_ZOCALO_CALLBACK_LOGGER)
        self._sample_id: int | None = None
        self._descriptor: str | None = None
        self._run_id: str | None = None

        # Record 'sample loaded' if document successfully stops
        self.record_loaded_on_success = record_loaded_on_success

    def activity_gated_start(self, doc: RunStart):
        if not self._sample_id and self.active:
            sample_id = doc.get("metadata", {}).get("sample_id")
            self.log.info(f"Recording sample ID at run start {sample_id}")
            self._sample_id = sample_id
            self._run_id = self.activity_uid

    def activity_gated_stop(self, doc: RunStop) -> RunStop:
        if self._run_id == doc.get("run_start"):
            expeye = ExpeyeInteraction()
            if doc["exit_status"] != "success":
                exception_type, message = SampleException.type_and_message_from_reason(
                    doc.get("reason", "")
                )
                self.log.info(
                    f"Sample handling callback intercepted exception of type {exception_type}: {message}"
                )
                self._record_exception(exception_type, expeye)

            elif self.record_loaded_on_success:
                self._record_loaded(expeye)

            self._sample_id = None
            self._run_id = None

        return doc

    def _record_exception(self, exception_type: str, expeye: ExpeyeInteraction):
        assert self._sample_id, "Unable to record exception due to no sample ID"
        sample_status = self._decode_sample_status(exception_type)
        expeye.update_sample_status(self._sample_id, sample_status)

    def _decode_sample_status(self, exception_type: str) -> BLSampleStatus:
        match exception_type:
            case SampleException.__name__ | CrystalNotFoundException.__name__:
                return BLSampleStatus.ERROR_SAMPLE
        return BLSampleStatus.ERROR_BEAMLINE

    def _record_loaded(self, expeye: ExpeyeInteraction):
        assert self._sample_id, "Unable to record loaded state due to no sample ID"
        expeye.update_sample_status(self._sample_id, BLSampleStatus.LOADED)



from __future__ import annotations

import pydantic
from dodal.devices.aperturescatterguard import (
    ApertureScatterguard,
)
from dodal.devices.attenuator.attenuator import BinaryFilterAttenuator
from dodal.devices.backlight import Backlight
from dodal.devices.dcm import DCM
from dodal.devices.eiger import EigerDetector
from dodal.devices.fast_grid_scan import (
    PandAFastGridScan,
    ZebraFastGridScan,
)
from dodal.devices.flux import Flux
from dodal.devices.robot import BartRobot
from dodal.devices.s4_slit_gaps import S4SlitGaps
from dodal.devices.smargon import Smargon
from dodal.devices.synchrotron import Synchrotron
from dodal.devices.undulator import Undulator
from dodal.devices.xbpm_feedback import XBPMFeedback
from dodal.devices.zebra.zebra import Zebra
from dodal.devices.zebra.zebra_controlled_shutter import ZebraShutter
from dodal.devices.zocalo import ZocaloResults
from ophyd_async.fastcs.panda import HDFPanda


@pydantic.dataclasses.dataclass(config={"arbitrary_types_allowed": True})
class HyperionFlyScanXRayCentreComposite:
    """All devices which are directly or indirectly required by this plan"""

    aperture_scatterguard: ApertureScatterguard
    attenuator: BinaryFilterAttenuator
    backlight: Backlight
    dcm: DCM
    eiger: EigerDetector
    zebra_fast_grid_scan: ZebraFastGridScan
    flux: Flux
    s4_slit_gaps: S4SlitGaps
    smargon: Smargon
    undulator: Undulator
    synchrotron: Synchrotron
    xbpm_feedback: XBPMFeedback
    zebra: Zebra
    zocalo: ZocaloResults
    panda: HDFPanda
    panda_fast_grid_scan: PandAFastGridScan
    robot: BartRobot
    sample_shutter: ZebraShutter



from unittest.mock import MagicMock, patch

from bluesky.plan_stubs import null
from bluesky.run_engine import RunEngine
from ophyd_async.fastcs.panda import HDFPanda

from mx_bluesky.common.device_setup_plans.setup_panda import load_panda_from_yaml


def get_test_plan(*args):
    yield from null()
    return "retrieved_settings"


@patch("mx_bluesky.common.device_setup_plans.setup_panda.YamlSettingsProvider")
@patch("mx_bluesky.common.device_setup_plans.setup_panda.retrieve_settings")
@patch("mx_bluesky.common.device_setup_plans.setup_panda.apply_panda_settings")
def test_load_panda_from_yaml(
    mock_apply_panda_settings: MagicMock,
    mock_retrieve_settings: MagicMock,
    mock_settings_provider: MagicMock,
    panda: HDFPanda,
    tmpdir,
    RE: RunEngine,
):
    test_file = "test"
    mock_settings_provider.return_value = (mock_settings_return := MagicMock())
    mock_retrieve_settings.side_effect = get_test_plan

    RE(
        load_panda_from_yaml(
            tmpdir,
            test_file,
            panda,
        )
    )

    mock_settings_provider.assert_called_once_with(tmpdir)
    mock_retrieve_settings.assert_called_once_with(
        mock_settings_return, test_file, panda
    )
    mock_apply_panda_settings.assert_called_once_with("retrieved_settings")



from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar

from mx_bluesky.common.external_interaction.callbacks.common.plan_reactive_callback import (
    PlanReactiveCallback,
)
from mx_bluesky.common.external_interaction.nexus.nexus_utils import (
    create_beam_and_attenuator_parameters,
    vds_type_based_on_bit_depth,
)
from mx_bluesky.common.external_interaction.nexus.write_nexus import NexusWriter
from mx_bluesky.common.parameters.constants import DocDescriptorNames, PlanNameConstants
from mx_bluesky.common.parameters.gridscan import (
    SpecifiedThreeDGridScan,
)
from mx_bluesky.common.utils.log import NEXUS_LOGGER

if TYPE_CHECKING:
    from event_model.documents import Event, EventDescriptor, RunStart

T = TypeVar("T", bound="SpecifiedThreeDGridScan")


class GridscanNexusFileCallback(PlanReactiveCallback):
    """Callback class to handle the creation of Nexus files based on experiment \
    parameters. Initialises on recieving a 'start' document for the \
    'run_gridscan_move_and_tidy' sub plan, which must also contain the run parameters, \
    as metadata under the 'hyperion_internal_parameters' key. Actually writes the \
    nexus files on updates the timestamps on recieving the 'ispyb_reading_hardware' event \
    document, and finalises the files on getting a 'stop' document for the whole run.

    To use, subscribe the Bluesky RunEngine to an instance of this class.
    E.g.:
        nexus_file_handler_callback = NexusFileCallback(parameters)
        RE.subscribe(nexus_file_handler_callback)
    Or decorate a plan using bluesky.preprocessors.subs_decorator.

    See: https://blueskyproject.io/bluesky/callbacks.html#ways-to-invoke-callbacks
    """

    def __init__(self, param_type: type[T]) -> None:
        super().__init__(NEXUS_LOGGER)
        self.param_type = param_type
        self.run_start_uid: str | None = None
        self.nexus_writer_1: NexusWriter | None = None
        self.nexus_writer_2: NexusWriter | None = None
        self.descriptors: dict[str, EventDescriptor] = {}
        self.log = NEXUS_LOGGER

    def activity_gated_start(self, doc: RunStart):
        if doc.get("subplan_name") == PlanNameConstants.GRIDSCAN_OUTER:
            mx_bluesky_parameters = doc.get("mx_bluesky_parameters")
            assert isinstance(mx_bluesky_parameters, str)
            NEXUS_LOGGER.info(
                f"Nexus writer received start document with experiment parameters {mx_bluesky_parameters}"
            )
            parameters = self.param_type.model_validate_json(mx_bluesky_parameters)
            d_size = parameters.detector_params.detector_size_constants.det_size_pixels
            grid_n_img_1 = parameters.scan_indices[1]
            grid_n_img_2 = parameters.num_images - grid_n_img_1
            data_shape_1 = (grid_n_img_1, d_size.width, d_size.height)
            data_shape_2 = (grid_n_img_2, d_size.width, d_size.height)
            run_number_2 = parameters.detector_params.run_number + 1
            self.nexus_writer_1 = NexusWriter(
                parameters, data_shape_1, parameters.scan_points_first_grid
            )
            self.nexus_writer_2 = NexusWriter(
                parameters,
                data_shape_2,
                parameters.scan_points_second_grid,
                run_number=run_number_2,
                vds_start_index=parameters.scan_indices[1],
                omega_start_deg=90,
            )
            self.run_start_uid = doc.get("uid")

    def activity_gated_descriptor(self, doc: EventDescriptor):
        self.descriptors[doc["uid"]] = doc

    def activity_gated_event(self, doc: Event) -> Event | None:
        event_descriptor = self.descriptors.get(doc["descriptor"])
        assert event_descriptor is not None
        if event_descriptor.get("name") == DocDescriptorNames.HARDWARE_READ_DURING:
            data = doc["data"]
            for nexus_writer in [self.nexus_writer_1, self.nexus_writer_2]:
                assert nexus_writer, "Nexus callback did not receive start doc"
                (
                    nexus_writer.beam,
                    nexus_writer.attenuator,
                ) = create_beam_and_attenuator_parameters(
                    data["dcm-energy_in_kev"],
                    data["flux-flux_reading"],
                    data["attenuator-actual_transmission"],
                )
                vds_data_type = vds_type_based_on_bit_depth(
                    doc["data"]["eiger_bit_depth"]
                )
                nexus_writer.create_nexus_file(vds_data_type)
                NEXUS_LOGGER.info(f"Nexus file created at {nexus_writer.data_filename}")

        return super().activity_gated_event(doc)



from unittest.mock import MagicMock
from uuid import uuid4

import numpy as np
import pytest
from daq_config_server.client import ConfigServer

from mx_bluesky.hyperion.external_interaction.config_server import HyperionFeatureFlags


@pytest.fixture
def config_service():
    return HyperionFeatureFlags.get_config_server()


@pytest.mark.s03
def test_get_beamline_params(config_service: ConfigServer):
    resp = config_service.get_beamline_param("miniap_x_SMALL_APERTURE")
    assert isinstance(resp, float)
    assert np.isclose(resp, 2.43)


@pytest.mark.s03
def test_get_feature_flag(config_service: ConfigServer):
    resp = config_service.get_feature_flag("set_stub_offsets")
    assert isinstance(resp, bool)


@pytest.mark.s03
def test_get_feature_flags(config_service: ConfigServer):
    features = config_service.best_effort_get_all_feature_flags()
    assert len(features.keys()) == 3


@pytest.mark.s03
def test_nonsense_feature_flag_fails_with_normal_call(config_service: ConfigServer):
    with pytest.raises(AssertionError):
        _ = config_service.get_feature_flag(str(uuid4()))


@pytest.mark.s03
def test_best_effort_gracefully_fails_with_nonsense(config_service: ConfigServer):
    resp = config_service.best_effort_get_feature_flag(str(uuid4()))
    assert resp is None


@pytest.mark.s03
def test_best_effort_gracefully_fails_if_service_down(config_service: ConfigServer):
    log_mock = MagicMock()
    config_service = ConfigServer("http://not_real_address", log_mock)
    resp = config_service.best_effort_get_feature_flag("set_stub_offsets")
    assert resp is None
    log_mock.error.assert_called_with(
        "Encountered an error reading from the config service.", exc_info=True
    )



from unittest.mock import patch

from mx_bluesky.beamlines.i24.serial.write_nexus import call_nexgen


@patch("mx_bluesky.beamlines.i24.serial.write_nexus.bps.sleep")
@patch("mx_bluesky.beamlines.i24.serial.write_nexus.SSX_LOGGER")
@patch("mx_bluesky.beamlines.i24.serial.write_nexus.caget")
@patch("mx_bluesky.beamlines.i24.serial.write_nexus.cagetstring")
@patch("mx_bluesky.beamlines.i24.serial.write_nexus.pathlib.Path.read_text")
@patch("mx_bluesky.beamlines.i24.serial.write_nexus.pathlib.Path.exists")
def test_call_nexgen_for_extruder(
    fake_path,
    fake_read_text,
    fake_caget_str,
    fake_caget,
    fake_log,
    sleep,
    dummy_params_ex,
    RE,
):
    fake_caget_str.return_value = f"{dummy_params_ex.filename}_5001"
    fake_caget.return_value = 32
    fake_path.return_value = True
    fake_read_text.return_value = ""
    with patch("mx_bluesky.beamlines.i24.serial.write_nexus.requests") as patch_request:
        RE(call_nexgen(None, dummy_params_ex, 0.6, (1000, 1200)))
        patch_request.post.assert_called_once()

    assert patch_request.post.call_args.kwargs["json"]["expt_type"] == "extruder"


@patch("mx_bluesky.beamlines.i24.serial.write_nexus.bps.sleep")
@patch("mx_bluesky.beamlines.i24.serial.write_nexus.SSX_LOGGER")
@patch("mx_bluesky.beamlines.i24.serial.write_nexus.caget")
@patch("mx_bluesky.beamlines.i24.serial.write_nexus.cagetstring")
@patch("mx_bluesky.beamlines.i24.serial.write_nexus.pathlib.Path.read_text")
@patch("mx_bluesky.beamlines.i24.serial.write_nexus.pathlib.Path.exists")
def test_call_nexgen_for_fixed_target(
    fake_path,
    fake_read_text,
    fake_caget_str,
    fake_caget,
    fake_log,
    sleep,
    dummy_params_without_pp,
    RE,
):
    expected_filename = f"{dummy_params_without_pp.filename}_5002"
    fake_caget_str.return_value = expected_filename
    fake_caget.return_value = 32
    fake_path.return_value = True
    fake_read_text.return_value = ""
    with patch("mx_bluesky.beamlines.i24.serial.write_nexus.requests") as patch_request:
        RE(call_nexgen(None, dummy_params_without_pp, 0.6, (1000, 1200)))
        patch_request.post.assert_called_once()

    assert patch_request.post.call_args.kwargs["json"]["expt_type"] == "fixed-target"
    assert patch_request.post.call_args.kwargs["json"]["filename"] == expected_filename



"""Callbacks which can be subscribed to by the Bluesky RunEngine in order to perform
external interactions in response to the 'documents' emitted when events occur in the
execution of an experimental plan.

Callbacks used for the Hyperion fast grid scan are prefixed with 'FGS'.
"""



import functools
import logging
import logging.config
import time
from os import environ
from pathlib import Path

import bluesky.plan_stubs as bps
from bluesky.log import logger as bluesky_logger
from bluesky.utils import MsgGenerator
from dodal.log import DEFAULT_GRAYLOG_PORT, ophyd_async_logger
from dodal.log import LOGGER as dodal_logger

from mx_bluesky.common.utils.log import do_default_logging_setup

VISIT_PATH = Path("/dls_sw/i24/etc/ssx_current_visit.txt")


# Logging set up
SSX_LOGGER = logging.getLogger("I24serial")
SSX_LOGGER.addHandler(logging.NullHandler())
SSX_LOGGER.parent = dodal_logger


logging_config = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "class": "logging.Formatter",
            "format": "%(message)s",
        }
    },
    "handlers": {
        "console": {
            "level": "INFO",
            "class": "logging.StreamHandler",
            "formatter": "default",
            "stream": "ext://sys.stdout",
        }
    },
    "loggers": {
        "I24serial": {
            "handlers": ["console"],
            "level": "DEBUG",
            "propagate": False,
        }
    },
}

logging.config.dictConfig(logging_config)


def _read_visit_directory_from_file() -> Path:
    with open(VISIT_PATH) as f:
        visit = f.readline().rstrip()
    return Path(visit)


def _get_logging_file_path() -> Path:
    """Get the path to write the serial experiment specific log file to.
    If on a beamline, this will be written to the tmp folder in the current visit.
    Returns:
        logging_path (Path): Path to the log file for the file handler to write to.
    """
    beamline: str | None = environ.get("BEAMLINE")
    logging_path: Path

    if beamline:
        logging_path = _read_visit_directory_from_file() / "tmp/serial/logs"
    else:
        logging_path = Path("./tmp/logs/")

    Path(logging_path).mkdir(parents=True, exist_ok=True)
    return logging_path


def _integrate_bluesky_logs(parent_logger: logging.Logger):
    # Integrate only bluesky and ophyd_async logger
    for log in [bluesky_logger, ophyd_async_logger]:
        log.parent = parent_logger
        log.setLevel(logging.DEBUG)


def config(
    logfile: str | None = None,
    write_mode: str = "a",
    delayed: bool = False,
    dev_mode: bool = False,
):
    """
    Configure the logging.

    Args:
        logfile (str, optional): Filename for logfile. If passed, create a file handler\
            for the logger to write to file the log output. Defaults to None.
        write_mode (str, optional): String indicating writing mode for the output \
            .log file. Defaults to "a".
        dev_mode (bool, optional): If true, will log to graylog on localhost instead \
            of production. Defaults to False.
    """
    do_default_logging_setup(
        "mx-bluesky.log",
        DEFAULT_GRAYLOG_PORT,
        dev_mode=dev_mode,
        integrate_all_logs=False,
    )
    # Remove dodal StreamHandler to avoid duplication of messages above debug
    dodal_logger.removeHandler(dodal_logger.handlers[0])
    _integrate_bluesky_logs(dodal_logger)

    if logfile:
        logs = _get_logging_file_path() / logfile
        fileFormatter = logging.Formatter(
            "%(asctime)s %(levelname)s: \t(%(name)s) %(message)s",
            datefmt="%d-%m-%Y %I:%M:%S",
        )
        FH = logging.FileHandler(logs, mode=write_mode, encoding="utf-8", delay=delayed)
        FH.setLevel(logging.DEBUG)
        FH.setFormatter(fileFormatter)
        SSX_LOGGER.addHandler(FH)


def log_on_entry(func):
    @functools.wraps(func)
    def decorator(*args, **kwargs):
        name = func.__name__
        SSX_LOGGER.debug(f"Running {name} ")
        return func(*args, **kwargs)

    return decorator


def setup_collection_logs(expt: str, dev_mode: bool = False) -> MsgGenerator:
    """A small plan to set up the logging from blueapi on start up as we're running \
        on procserv.
        This setup will likely change once the beamline has a cluster.
    """
    if (
        expt == "Serial Fixed"
    ):  # SSXType.FIXED: See https://github.com/DiamondLightSource/mx-bluesky/issues/608
        logfile = time.strftime("i24fixedtarget_%d%B%y.log").lower()
    else:
        logfile = time.strftime("i24extruder_%d%B%y.log").lower()

    config(logfile, dev_mode=dev_mode)
    yield from bps.null()


def clean_up_log_config_at_end() -> MsgGenerator:
    """A small plan for blueapi to tidy up logging configuration."""
    # See https://github.com/DiamondLightSource/mx-bluesky/issues/609
    for handler in SSX_LOGGER.handlers:
        SSX_LOGGER.removeHandler(handler)
    for handler in dodal_logger.handlers:
        dodal_logger.removeHandler(handler)
    yield from bps.null()



from __future__ import annotations

import math
from typing import TYPE_CHECKING

import bluesky.plan_stubs as bps
import numpy as np
import pydantic
from blueapi.core import BlueskyContext
from dodal.devices.backlight import Backlight
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.oav.pin_image_recognition import PinTipDetection
from dodal.devices.oav.pin_image_recognition.utils import NONE_VALUE
from dodal.devices.oav.utils import PinNotFoundException, wait_for_tip_to_be_found
from dodal.devices.smargon import Smargon

from mx_bluesky.common.utils.context import device_composite_from_context
from mx_bluesky.common.utils.exceptions import catch_exception_and_warn
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.hyperion.device_setup_plans.setup_oav import (
    pre_centring_setup_oav,
)
from mx_bluesky.hyperion.parameters.constants import CONST

if TYPE_CHECKING:
    from dodal.devices.oav.oav_parameters import OAVParameters


@pydantic.dataclasses.dataclass(config={"arbitrary_types_allowed": True})
class OavGridDetectionComposite:
    """All devices which are directly or indirectly required by this plan"""

    backlight: Backlight
    oav: OAV
    smargon: Smargon
    pin_tip_detection: PinTipDetection


def create_devices(context: BlueskyContext) -> OavGridDetectionComposite:
    return device_composite_from_context(context, OavGridDetectionComposite)


def get_min_and_max_y_of_pin(
    top: np.ndarray, bottom: np.ndarray, full_image_height_px: int
) -> tuple[int, int]:
    """Gives the minimum and maximum y that would cover the whole pin.

    First filters out where no edge was found or the edge covers the full image.
    If this results in no edges found then returns a min/max that covers the full image
    """
    filtered_top = top[np.where((top != 0) & (top != NONE_VALUE))]
    min_y = min(filtered_top) if len(filtered_top) else 0
    filtered_bottom = bottom[
        np.where((bottom != full_image_height_px) & (bottom != NONE_VALUE))
    ]
    max_y = max(filtered_bottom) if len(filtered_bottom) else full_image_height_px
    return min_y, max_y


def grid_detection_plan(
    composite: OavGridDetectionComposite,
    parameters: OAVParameters,
    snapshot_template: str,
    snapshot_dir: str,
    grid_width_microns: float,
    box_size_um: float,
):
    """
    Creates the parameters for two grids that are 90 degrees from each other and
    encompass the whole of the sample as it appears in the OAV.

    Args:
        composite (OavGridDetectionComposite): Composite containing devices for doing a grid detection.
        parameters (OAVParameters): Object containing parameters for setting up the OAV
        snapshot_template (str): A template for the name of the snapshots, expected to be filled in with an angle
        snapshot_dir (str): The location to save snapshots
        grid_width_microns (float): The width of the grid to scan in microns
        box_size_um (float): The size of each box of the grid in microns
    """
    oav: OAV = composite.oav
    smargon: Smargon = composite.smargon
    pin_tip_detection: PinTipDetection = composite.pin_tip_detection

    LOGGER.info("OAV Centring: Starting grid detection centring")

    yield from bps.wait()

    # Set relevant PVs to whatever the config dictates.
    yield from pre_centring_setup_oav(oav, parameters, pin_tip_detection)

    LOGGER.info("OAV Centring: Camera set up")

    microns_per_pixel_x = yield from bps.rd(oav.microns_per_pixel_x)
    microns_per_pixel_y = yield from bps.rd(oav.microns_per_pixel_y)

    box_size_x_pixels = box_size_um / microns_per_pixel_x
    box_size_y_pixels = box_size_um / microns_per_pixel_y

    grid_width_pixels = int(grid_width_microns / microns_per_pixel_x)

    # The FGS uses -90 so we need to match it
    for angle in [0, -90]:
        yield from bps.mv(smargon.omega, angle)
        # need to wait for the OAV image to update
        # See #673 for improvements
        yield from bps.sleep(CONST.HARDWARE.OAV_REFRESH_DELAY)

        tip_x_px, tip_y_px = yield from catch_exception_and_warn(
            PinNotFoundException, wait_for_tip_to_be_found, pin_tip_detection
        )

        LOGGER.info(f"Tip is at x,y: {tip_x_px},{tip_y_px}")

        top_edge = np.array((yield from bps.rd(pin_tip_detection.triggered_top_edge)))
        bottom_edge = np.array(
            (yield from bps.rd(pin_tip_detection.triggered_bottom_edge))
        )

        full_image_height_px = yield from bps.rd(oav.cam.array_size_y)

        # only use the area from the start of the pin onwards
        top_edge = top_edge[tip_x_px : tip_x_px + grid_width_pixels]
        bottom_edge = bottom_edge[tip_x_px : tip_x_px + grid_width_pixels]
        LOGGER.info(f"OAV Edge detection top: {list(top_edge)}")
        LOGGER.info(f"OAV Edge detection bottom: {list(bottom_edge)}")

        min_y, max_y = (
            float(n)
            for n in get_min_and_max_y_of_pin(
                top_edge, bottom_edge, full_image_height_px
            )
        )

        grid_height_px: float = max_y - min_y

        y_steps: int = math.ceil(grid_height_px / box_size_y_pixels)

        # Panda not configured to run a half complete snake so enforce even rows on first grid
        # See https://github.com/DiamondLightSource/hyperion/wiki/PandA-constant%E2%80%90motion-scanning#motion-program-summary
        if y_steps % 2 and angle == 0:
            LOGGER.debug(
                f"Forcing number of rows in first grid to be even: Adding an extra row onto bottom of first grid and shifting grid upwards by {box_size_y_pixels / 2}"
            )
            y_steps += 1
            min_y -= box_size_y_pixels / 2
            max_y += box_size_y_pixels / 2
            grid_height_px += box_size_y_pixels

        LOGGER.info(f"Drawing snapshot {grid_width_pixels} by {grid_height_px}")

        x_steps = math.ceil(grid_width_pixels / box_size_x_pixels)

        upper_left = (tip_x_px, min_y)

        yield from bps.abs_set(oav.grid_snapshot.top_left_x, upper_left[0])
        yield from bps.abs_set(oav.grid_snapshot.top_left_y, upper_left[1])
        yield from bps.abs_set(oav.grid_snapshot.box_width, box_size_x_pixels)
        yield from bps.abs_set(oav.grid_snapshot.num_boxes_x, x_steps)
        yield from bps.abs_set(oav.grid_snapshot.num_boxes_y, y_steps)

        snapshot_filename = snapshot_template.format(angle=abs(angle))

        yield from bps.abs_set(oav.grid_snapshot.filename, snapshot_filename)
        yield from bps.abs_set(oav.grid_snapshot.directory, snapshot_dir)
        yield from bps.trigger(oav.grid_snapshot, wait=True)
        yield from bps.create(CONST.DESCRIPTORS.OAV_GRID_SNAPSHOT_TRIGGERED)

        yield from bps.read(oav)
        yield from bps.read(smargon)
        yield from bps.save()

        LOGGER.info(
            f"Grid calculated at {angle}: {x_steps} by {y_steps} steps starting at {upper_left}px"
        )



from blueapi.core import BlueskyContext
from dodal.utils import get_beamline_based_on_environment_variable

import mx_bluesky.hyperion.experiment_plans as hyperion_plans
from mx_bluesky.common.utils.log import LOGGER


def setup_context(wait_for_connection: bool = True) -> BlueskyContext:
    context = BlueskyContext()
    context.with_plan_module(hyperion_plans)

    context.with_dodal_module(
        get_beamline_based_on_environment_variable(),
        wait_for_connection=wait_for_connection,
    )

    LOGGER.info(f"Plans found in context: {context.plan_functions.keys()}")

    return context



from unittest.mock import patch

import pytest
from bluesky.run_engine import RunEngine
from dodal.devices.i24.i24_detector_motion import DetectorMotion
from ophyd_async.testing import set_mock_value

from mx_bluesky.beamlines.i24.serial.parameters.constants import SSXType
from mx_bluesky.beamlines.i24.serial.setup_beamline import Eiger, Pilatus
from mx_bluesky.beamlines.i24.serial.setup_beamline.setup_detector import (
    DetRequest,
    _get_requested_detector,
    get_detector_type,
    setup_detector_stage,
)


def test_get_detector_type(RE, detector_stage: DetectorMotion):
    set_mock_value(detector_stage.y.user_readback, -22)
    det_type = RE(get_detector_type(detector_stage)).plan_result
    assert det_type.name == "eiger"


def test_get_detector_type_finds_pilatus(RE, detector_stage: DetectorMotion):
    set_mock_value(detector_stage.y.user_readback, 566)
    det_type = RE(get_detector_type(detector_stage)).plan_result
    assert det_type.name == "pilatus"


@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_detector.caget")
def test_get_requested_detector(fake_caget):
    fake_caget.return_value = "pilatus"
    assert _get_requested_detector("some_pv") == Pilatus.name

    fake_caget.return_value = "0"
    assert _get_requested_detector("some_pv") == Eiger.name


@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_detector.caget")
def test_get_requested_detector_raises_error_for_invalid_value(fake_caget):
    fake_caget.return_value = "something"
    with pytest.raises(ValueError):
        _get_requested_detector("some_pv")


@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_detector.caget")
async def test_setup_detector_stage(
    fake_caget, detector_stage: DetectorMotion, RE: RunEngine
):
    fake_caget.return_value = DetRequest.eiger.value
    RE(setup_detector_stage(SSXType.FIXED, detector_stage))
    assert await detector_stage.y.user_setpoint.get_value() == Eiger.det_y_target

    fake_caget.return_value = DetRequest.pilatus.value
    RE(setup_detector_stage(SSXType.EXTRUDER, detector_stage))
    assert await detector_stage.y.user_setpoint.get_value() == Pilatus.det_y_target



from __future__ import annotations

from collections.abc import Generator
from datetime import datetime
from pathlib import Path
from typing import cast

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
import pydantic
from blueapi.core import BlueskyContext
from bluesky.utils import Msg
from dodal.devices.aperturescatterguard import ApertureScatterguard, ApertureValue
from dodal.devices.attenuator.attenuator import BinaryFilterAttenuator
from dodal.devices.dcm import DCM
from dodal.devices.focusing_mirror import FocusingMirrorWithStripes, MirrorVoltages
from dodal.devices.motors import XYZPositioner
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.robot import BartRobot, SampleLocation
from dodal.devices.smargon import Smargon, StubPosition
from dodal.devices.thawer import Thawer
from dodal.devices.undulator_dcm import UndulatorDCM
from dodal.devices.webcam import Webcam
from dodal.devices.xbpm_feedback import XBPMFeedback
from dodal.plan_stubs.motor_utils import MoveTooLarge, home_and_reset_wrapper

from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.hyperion.experiment_plans.set_energy_plan import (
    SetEnergyComposite,
    set_energy_plan,
)
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.robot_load import RobotLoadAndEnergyChange


@pydantic.dataclasses.dataclass(config={"arbitrary_types_allowed": True})
class RobotLoadAndEnergyChangeComposite:
    # SetEnergyComposite fields
    vfm: FocusingMirrorWithStripes
    mirror_voltages: MirrorVoltages
    dcm: DCM
    undulator_dcm: UndulatorDCM
    xbpm_feedback: XBPMFeedback
    attenuator: BinaryFilterAttenuator

    # RobotLoad fields
    robot: BartRobot
    webcam: Webcam
    lower_gonio: XYZPositioner
    thawer: Thawer
    oav: OAV
    smargon: Smargon
    aperture_scatterguard: ApertureScatterguard


def create_devices(context: BlueskyContext) -> RobotLoadAndEnergyChangeComposite:
    from mx_bluesky.common.utils.context import device_composite_from_context

    return device_composite_from_context(context, RobotLoadAndEnergyChangeComposite)


def wait_for_smargon_not_disabled(smargon: Smargon, timeout=60):
    """Waits for the smargon disabled flag to go low. The robot hardware is responsible
    for setting this to low when it is safe to move. It does this through a physical
    connection between the robot and the smargon.
    """
    LOGGER.info("Waiting for smargon enabled")
    SLEEP_PER_CHECK = 0.1
    times_to_check = int(timeout / SLEEP_PER_CHECK)
    for _ in range(times_to_check):
        smargon_disabled = yield from bps.rd(smargon.disabled)
        if not smargon_disabled:
            LOGGER.info("Smargon now enabled")
            return
        yield from bps.sleep(SLEEP_PER_CHECK)
    raise TimeoutError(
        "Timed out waiting for smargon to become enabled after robot load"
    )


def take_robot_snapshots(oav: OAV, webcam: Webcam, directory: Path):
    time_now = datetime.now()
    snapshot_format = f"{time_now.strftime('%H%M%S')}_{{device}}_after_load"
    for device in [oav.snapshot, webcam]:
        yield from bps.abs_set(
            device.filename, snapshot_format.format(device=device.name)
        )
        yield from bps.abs_set(device.directory, str(directory))
        # Note: should be able to use `wait=True` after https://github.com/bluesky/bluesky/issues/1795
        yield from bps.trigger(device, group="snapshots")
        yield from bps.wait("snapshots")


def prepare_for_robot_load(
    aperture_scatterguard: ApertureScatterguard, smargon: Smargon
):
    yield from bps.abs_set(
        aperture_scatterguard, ApertureValue.OUT_OF_BEAM, group="prepare_robot_load"
    )

    yield from bps.mv(smargon.stub_offsets, StubPosition.RESET_TO_ROBOT_LOAD)

    # fmt: off
    yield from bps.mv(
        smargon.x, 0,
        smargon.y, 0,
        smargon.z, 0,
        smargon.omega, 0,
        smargon.chi, 0,
        smargon.phi, 0
    )
    # fmt: on

    yield from bps.wait("prepare_robot_load")


def do_robot_load(
    composite: RobotLoadAndEnergyChangeComposite,
    sample_location: SampleLocation,
    demand_energy_ev: float | None,
    thawing_time: float,
):
    yield from bps.abs_set(
        composite.robot,
        sample_location,
        group="robot_load",
    )

    yield from set_energy_plan(demand_energy_ev, cast(SetEnergyComposite, composite))

    yield from bps.wait("robot_load")

    yield from bps.abs_set(
        composite.thawer.thaw_for_time_s,
        thawing_time,
        group="thawing_finished",
    )
    yield from wait_for_smargon_not_disabled(composite.smargon)


def raise_exception_if_moved_out_of_cryojet(exception):
    yield from bps.null()
    if isinstance(exception, MoveTooLarge):
        raise Exception(
            f"Moving {exception.axis} back to {exception.position} after \
                        robot load would move it out of the cryojet. The max safe \
                        distance is {exception.maximum_move}"
        )


def pin_already_loaded(
    robot: BartRobot, sample_location: SampleLocation
) -> Generator[Msg, None, bool]:
    current_puck = yield from bps.rd(robot.current_puck)
    current_pin = yield from bps.rd(robot.current_pin)
    return (
        int(current_puck) == sample_location.puck
        and int(current_pin) == sample_location.pin
    )


def robot_load_and_snapshots(
    composite: RobotLoadAndEnergyChangeComposite,
    location: SampleLocation,
    snapshot_directory: Path,
    thawing_time: float,
    demand_energy_ev: float | None,
):
    robot_load_plan = do_robot_load(
        composite,
        location,
        demand_energy_ev,
        thawing_time,
    )

    # The lower gonio must be in the correct position for the robot load and we
    # want to put it back afterwards. Note we don't wait the robot is interlocked
    # to the lower gonio and the  move is quicker than the robot takes to get to the
    # load position.
    yield from bpp.contingency_wrapper(
        home_and_reset_wrapper(
            robot_load_plan,
            composite.lower_gonio,
            BartRobot.LOAD_TOLERANCE_MM,
            CONST.HARDWARE.CRYOJET_MARGIN_MM,
            "lower_gonio",
            wait_for_all=False,
        ),
        except_plan=raise_exception_if_moved_out_of_cryojet,
    )

    yield from take_robot_snapshots(composite.oav, composite.webcam, snapshot_directory)

    yield from bps.create(name=CONST.DESCRIPTORS.ROBOT_LOAD)
    yield from bps.read(composite.robot.barcode)
    yield from bps.read(composite.oav.snapshot)
    yield from bps.read(composite.webcam)
    yield from bps.save()

    yield from bps.wait("reset-lower_gonio")


def robot_load_and_change_energy_plan(
    composite: RobotLoadAndEnergyChangeComposite,
    params: RobotLoadAndEnergyChange,
):
    assert params.sample_puck is not None
    assert params.sample_pin is not None

    sample_location = SampleLocation(params.sample_puck, params.sample_pin)

    yield from prepare_for_robot_load(
        composite.aperture_scatterguard, composite.smargon
    )

    yield from bpp.set_run_key_wrapper(
        bpp.run_wrapper(
            robot_load_and_snapshots(
                composite,
                sample_location,
                params.snapshot_directory,
                params.thawing_time,
                params.demand_energy_ev,
            ),
            md={
                "subplan_name": CONST.PLAN.ROBOT_LOAD,
                "metadata": {
                    "visit": params.visit,
                    "sample_id": params.sample_id,
                    "sample_puck": sample_location.puck,
                    "sample_pin": sample_location.pin,
                },
                "activate_callbacks": [
                    "RobotLoadISPyBCallback",
                ],
            },
        ),
        CONST.PLAN.ROBOT_LOAD_AND_SNAPSHOTS,
    )



from unittest.mock import MagicMock

import pydantic
import pytest
from ophyd.device import Device

from mx_bluesky.common.utils.context import (
    device_composite_from_context,
    find_device_in_context,
)


class _DeviceType1(Device):
    pass


class _DeviceType2(Device):
    pass


def test_find_device_in_context():
    context = MagicMock()
    device = MagicMock(spec=Device)
    context.find_device.return_value = device

    found_device = find_device_in_context(context, "", expected_type=Device)  # type: ignore
    assert found_device == device


def find_device_in_context_with_wrong_type_raises_error():
    context = MagicMock()

    device = MagicMock(spec=_DeviceType1)
    context.find_device.return_value = device

    # Should not raise
    find_device_in_context(context, "", expected_type=_DeviceType1)

    with pytest.raises(ValueError):
        # Should raise
        find_device_in_context(context, "", expected_type=_DeviceType2)


def test_find_nonexistent_device_in_context_raises_error():
    context = MagicMock()
    context.find_device.return_value = None

    with pytest.raises(ValueError):
        find_device_in_context(context, "", Device)


def test_device_composite_from_context():
    context = MagicMock()

    @pydantic.dataclasses.dataclass(config={"arbitrary_types_allowed": True})
    class _Composite:
        device1: _DeviceType1
        device2: _DeviceType2

    device1_instance = MagicMock(spec=_DeviceType1)
    device2_instance = MagicMock(spec=_DeviceType2)

    context.find_device = lambda name: {
        "device1": device1_instance,
        "device2": device2_instance,
    }.get(name)

    composite = device_composite_from_context(context, _Composite)

    assert composite.device1 == device1_instance
    assert isinstance(composite.device1, _DeviceType1)

    assert composite.device2 == device2_instance
    assert isinstance(composite.device2, _DeviceType2)



import bluesky.plan_stubs as bps
import numpy
from dodal.devices.aperturescatterguard import ApertureScatterguard, ApertureValue
from dodal.devices.smargon import Smargon, StubPosition

from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.common.utils.tracing import TRACER
from mx_bluesky.common.xrc_result import XRayCentreResult
from mx_bluesky.hyperion.device_setup_plans.manipulate_sample import move_x_y_z
from mx_bluesky.hyperion.parameters.gridscan import HyperionSpecifiedThreeDGridScan


def change_aperture_then_move_to_xtal(
    best_hit: XRayCentreResult,
    smargon: Smargon,
    aperture_scatterguard: ApertureScatterguard,
    parameters: HyperionSpecifiedThreeDGridScan | None = None,
):
    """For the given x-ray centring result,
    * Change the aperture so that the beam size is comparable to the crystal size
    * Centre on the centre-of-mass
    * Reset the stub offsets if specified by params"""
    bounding_box_size = numpy.abs(
        best_hit.bounding_box_mm[1] - best_hit.bounding_box_mm[0]
    )
    yield from set_aperture_for_bbox_mm(
        aperture_scatterguard,
        bounding_box_size,
    )

    # once we have the results, go to the appropriate position
    LOGGER.info("Moving to centre of mass.")
    with TRACER.start_span("move_to_result"):
        x, y, z = best_hit.centre_of_mass_mm
        yield from move_x_y_z(smargon, x, y, z, wait=True)

    # TODO support for setting stub offsets in multipin
    # https://github.com/DiamondLightSource/mx-bluesky/issues/552
    if parameters and parameters.FGS_params.set_stub_offsets:
        LOGGER.info("Recentring smargon co-ordinate system to this point.")
        yield from bps.mv(smargon.stub_offsets, StubPosition.CURRENT_AS_CENTER)


def set_aperture_for_bbox_mm(
    aperture_device: ApertureScatterguard,
    bbox_size_mm: list[float] | numpy.ndarray,
):
    """Sets aperture size based on bbox_size.

    This function determines the aperture size needed to accommodate the bounding box
    of a crystal. The x-axis length of the bounding box is used, setting the aperture
    to Medium if this is less than 50um, and Large otherwise.

    Args:
        aperture_device: The aperture scatter guard device we are controlling.
        bbox_size_mm: The [x,y,z] lengths, in mm, of a bounding box
        containing a crystal. This describes (in no particular order):
        * The maximum width a crystal occupies
        * The maximum height a crystal occupies
        * The maximum depth a crystal occupies
        constructing a three dimensional cuboid, completely encapsulating the crystal.

    Yields:
        Iterator[MsgGenerator]
    """

    # bbox_size is [x,y,z], for i03 we only care about x
    new_selected_aperture = (
        ApertureValue.MEDIUM if bbox_size_mm[0] < 0.05 else ApertureValue.LARGE
    )
    LOGGER.info(
        f"Setting aperture to {new_selected_aperture} based on bounding box size {bbox_size_mm}."
    )

    yield from bps.abs_set(aperture_device, new_selected_aperture)



from __future__ import annotations

import os
from collections.abc import Iterator
from itertools import accumulate
from typing import Annotated, Any, Self

from annotated_types import Len
from dodal.devices.aperturescatterguard import ApertureValue
from dodal.devices.detector import DetectorParams
from dodal.devices.zebra.zebra import (
    RotationDirection,
)
from dodal.log import LOGGER
from pydantic import Field, field_validator, model_validator
from scanspec.core import AxesPoints
from scanspec.core import Path as ScanPath
from scanspec.specs import Line

from mx_bluesky.common.parameters.components import (
    DiffractionExperimentWithSample,
    IspybExperimentType,
    OptionalGonioAngleStarts,
    OptionalXyzStarts,
    RotationAxis,
    SplitScan,
    WithScan,
)
from mx_bluesky.hyperion.parameters.components import WithHyperionUDCFeatures
from mx_bluesky.hyperion.parameters.constants import (
    CONST,
    I03Constants,
)


class RotationScanPerSweep(OptionalGonioAngleStarts, OptionalXyzStarts):
    """
    Describes a rotation scan about the specified axis.

    Attributes:
        rotation_axis: The rotation axis, by default this is the omega axis
        omega_start_deg: The initial angle of the rotation in degrees (default 0)
        scan_width_deg: The sweep of the rotation in degrees, this must be positive (default 360)
        rotation_direction: Indicates the direction of rotation, if RotationDirection.POSITIVE
            the final angle is obtained by adding scan_width_deg, otherwise by subtraction (default NEGATIVE)
        nexus_vds_start_img: The frame number of the first frame captured during the rotation
    """

    omega_start_deg: float = Field(default=0)  # type: ignore
    rotation_axis: RotationAxis = Field(default=RotationAxis.OMEGA)
    scan_width_deg: float = Field(default=360, gt=0)
    rotation_direction: RotationDirection = Field(default=RotationDirection.NEGATIVE)
    nexus_vds_start_img: int = Field(default=0, ge=0)


class RotationExperiment(DiffractionExperimentWithSample, WithHyperionUDCFeatures):
    shutter_opening_time_s: float = Field(default=CONST.I03.SHUTTER_TIME_S)
    rotation_increment_deg: float = Field(default=0.1, gt=0)
    ispyb_experiment_type: IspybExperimentType = Field(
        default=IspybExperimentType.ROTATION
    )

    def _detector_params_impl(
        self, omega_start_deg: float, num_images_per_trigger: int, num_triggers: int
    ) -> DetectorParams:
        self.det_dist_to_beam_converter_path = (
            self.det_dist_to_beam_converter_path
            or CONST.PARAM.DETECTOR.BEAM_XY_LUT_PATH
        )
        optional_args = {}
        if self.run_number:
            optional_args["run_number"] = self.run_number
        assert self.detector_distance_mm is not None
        os.makedirs(self.storage_directory, exist_ok=True)
        return DetectorParams(
            detector_size_constants=I03Constants.DETECTOR,
            expected_energy_ev=self.demand_energy_ev,
            exposure_time_s=self.exposure_time_s,
            directory=self.storage_directory,
            prefix=self.file_name,
            detector_distance=self.detector_distance_mm,
            omega_start=omega_start_deg,
            omega_increment=self.rotation_increment_deg,
            num_images_per_trigger=num_images_per_trigger,
            num_triggers=num_triggers,
            use_roi_mode=False,
            det_dist_to_beam_converter_path=self.det_dist_to_beam_converter_path,
            **optional_args,
        )

    def _detector_params(self, omega_start_deg: float) -> DetectorParams:
        return self._detector_params_impl(omega_start_deg, self.num_images, 1)

    @field_validator("selected_aperture")
    @classmethod
    def _set_default_aperture_position(cls, aperture_position: ApertureValue | None):
        if not aperture_position:
            default_aperture = CONST.PARAM.ROTATION.DEFAULT_APERTURE_POSITION
            LOGGER.warning(
                f"No aperture position selected. Defaulting to {default_aperture}"
            )
            return default_aperture
        else:
            return aperture_position


class RotationScan(WithScan, RotationScanPerSweep, RotationExperiment):
    @property
    def detector_params(self):
        return self._detector_params(self.omega_start_deg)

    @property
    def scan_points(self) -> AxesPoints:
        """The scan points are defined in application space"""
        scan_spec = Line(
            axis="omega",
            start=self.omega_start_deg,
            stop=(
                self.omega_start_deg
                + (self.scan_width_deg - self.rotation_increment_deg)
            ),
            num=self.num_images,
        )
        scan_path = ScanPath(scan_spec.calculate())
        return scan_path.consume().midpoints

    @property
    def num_images(self) -> int:
        return int(self.scan_width_deg / self.rotation_increment_deg)


class MultiRotationScan(RotationExperiment, SplitScan):
    rotation_scans: Annotated[list[RotationScanPerSweep], Len(min_length=1)]

    def _single_rotation_scan(self, scan: RotationScanPerSweep) -> RotationScan:
        # self has everything from RotationExperiment
        allowed_keys = RotationScan.model_fields.keys()  # type: ignore # mypy doesn't recognise this as a property...
        params_dump = self.model_dump()
        # provided `scan` has everything from RotationScanPerSweep
        scan_dump = scan.model_dump()
        rotation_scan_kv_pairs = {
            k: v for k, v in (params_dump | scan_dump).items() if k in allowed_keys
        }
        # together they have everything for RotationScan
        rotation_scan = RotationScan(**rotation_scan_kv_pairs)
        return rotation_scan

    @model_validator(mode="after")
    @classmethod
    def correct_start_vds(cls, values: Any) -> Any:
        assert isinstance(values, MultiRotationScan)
        start_img = 0.0
        for scan in values.rotation_scans:
            scan.nexus_vds_start_img = int(start_img)
            start_img += scan.scan_width_deg / values.rotation_increment_deg
        return values

    @model_validator(mode="after")
    def _check_valid_for_single_arm_multiple_sweep(self) -> Self:
        if len(self.rotation_scans) > 0:
            scan_width = self.rotation_scans[0].scan_width_deg
            for scan in self.rotation_scans[1:]:
                assert scan.scan_width_deg == scan_width, (
                    "Sweeps with different numbers of frames are not supported."
                )

        return self

    @property
    def single_rotation_scans(self) -> Iterator[RotationScan]:
        for scan in self.rotation_scans:
            yield self._single_rotation_scan(scan)

    def _num_images_per_scan(self):
        return [
            int(scan.scan_width_deg / self.rotation_increment_deg)
            for scan in self.rotation_scans
        ]

    @property
    def num_images(self):
        return sum(self._num_images_per_scan())

    @property
    def scan_indices(self):
        return list(accumulate([0, *self._num_images_per_scan()]))

    @property
    def detector_params(self) -> DetectorParams:
        return self._detector_params_impl(
            self.rotation_scans[0].omega_start_deg,
            self._num_images_per_scan()[0],
            len(self._num_images_per_scan()),
        )

    @property
    def ispyb_params(self):  # pyright: ignore
        raise ValueError("Please get ispyb params from one of the individual scans")



from __future__ import annotations

from collections.abc import Callable, Sequence
from time import time
from typing import TYPE_CHECKING, Any, TypeVar

import numpy as np
from bluesky import preprocessors as bpp
from bluesky.utils import MsgGenerator
from dodal.devices.zocalo.zocalo_results import (
    ZOCALO_READING_PLAN_NAME,
    get_processing_results_from_event,
)

from mx_bluesky.common.external_interaction.callbacks.common.ispyb_callback_base import (
    BaseISPyBCallback,
)
from mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping import (
    populate_data_collection_group,
    populate_remaining_data_collection_info,
)
from mx_bluesky.common.external_interaction.callbacks.common.logging_callback import (
    format_doc_for_log,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_mapping import (
    construct_comment_for_gridscan,
    populate_xy_data_collection_info,
    populate_xz_data_collection_info,
)
from mx_bluesky.common.external_interaction.ispyb.data_model import (
    DataCollectionGridInfo,
    DataCollectionGroupInfo,
    DataCollectionInfo,
    DataCollectionPositionInfo,
    Orientation,
    ScanDataInfo,
)
from mx_bluesky.common.external_interaction.ispyb.ispyb_store import (
    IspybIds,
    StoreInIspyb,
)
from mx_bluesky.common.parameters.components import DiffractionExperimentWithSample
from mx_bluesky.common.parameters.constants import DocDescriptorNames, PlanNameConstants
from mx_bluesky.common.parameters.gridscan import (
    GridCommon,
)
from mx_bluesky.common.utils.exceptions import (
    ISPyBDepositionNotMade,
    SampleException,
)
from mx_bluesky.common.utils.log import ISPYB_ZOCALO_CALLBACK_LOGGER, set_dcgid_tag

if TYPE_CHECKING:
    from event_model import Event, RunStart, RunStop

T = TypeVar("T", bound="GridCommon")
ASSERT_START_BEFORE_EVENT_DOC_MESSAGE = f"No data collection group info - event document has been emitted before a {PlanNameConstants.GRID_DETECT_AND_DO_GRIDSCAN} start document"


def ispyb_activation_wrapper(plan_generator: MsgGenerator, parameters):
    return bpp.set_run_key_wrapper(
        bpp.run_wrapper(
            plan_generator,
            md={
                "activate_callbacks": ["GridscanISPyBCallback"],
                "subplan_name": PlanNameConstants.GRID_DETECT_AND_DO_GRIDSCAN,
                "mx_bluesky_parameters": parameters.model_dump_json(),
            },
        ),
        PlanNameConstants.ISPYB_ACTIVATION,
    )


class GridscanISPyBCallback(BaseISPyBCallback):
    """Callback class to handle the deposition of experiment parameters into the ISPyB
    database. Listens for 'event' and 'descriptor' documents. Creates the ISpyB entry on
    recieving an 'event' document for the 'ispyb_reading_hardware' event, and updates the
    deposition on recieving its final 'stop' document.

    To use, subscribe the Bluesky RunEngine to an instance of this class.
    E.g.:
        ispyb_handler_callback = FGSISPyBCallback(parameters)
        RE.subscribe(ispyb_handler_callback)
    Or decorate a plan using bluesky.preprocessors.subs_decorator.

    See: https://blueskyproject.io/bluesky/callbacks.html#ways-to-invoke-callbacks
    """

    def __init__(
        self,
        param_type: type[T],
        *,
        emit: Callable[..., Any] | None = None,
    ) -> None:
        super().__init__(emit=emit)
        self.ispyb: StoreInIspyb
        self.ispyb_ids: IspybIds = IspybIds()
        self.param_type = param_type
        self._start_of_fgs_uid: str | None = None
        self._processing_start_time: float | None = None
        self.data_collection_group_info: DataCollectionGroupInfo | None

    def activity_gated_start(self, doc: RunStart):
        if doc.get("subplan_name") == PlanNameConstants.DO_FGS:
            self._start_of_fgs_uid = doc.get("uid")
        if doc.get("subplan_name") == PlanNameConstants.GRID_DETECT_AND_DO_GRIDSCAN:
            self.uid_to_finalize_on = doc.get("uid")
            ISPYB_ZOCALO_CALLBACK_LOGGER.info(
                "ISPyB callback received start document with experiment parameters and "
                f"uid: {self.uid_to_finalize_on}"
            )
            mx_bluesky_parameters = doc.get("mx_bluesky_parameters")
            assert isinstance(mx_bluesky_parameters, str)
            self.params = self.param_type.model_validate_json(mx_bluesky_parameters)
            self.ispyb = StoreInIspyb(self.ispyb_config)
            self.data_collection_group_info = populate_data_collection_group(
                self.params
            )

            scan_data_infos = [
                ScanDataInfo(
                    data_collection_info=populate_remaining_data_collection_info(
                        "MX-Bluesky: Xray centring 1 -",
                        None,
                        populate_xy_data_collection_info(
                            self.params.detector_params,
                        ),
                        self.params,
                    ),
                ),
                ScanDataInfo(
                    data_collection_info=populate_remaining_data_collection_info(
                        "MX-Bluesky: Xray centring 2 -",
                        None,
                        populate_xz_data_collection_info(self.params.detector_params),
                        self.params,
                    )
                ),
            ]

            self.ispyb_ids = self.ispyb.begin_deposition(
                self.data_collection_group_info, scan_data_infos
            )
            set_dcgid_tag(self.ispyb_ids.data_collection_group_id)
        return super().activity_gated_start(doc)

    def activity_gated_event(self, doc: Event):
        assert self.data_collection_group_info, ASSERT_START_BEFORE_EVENT_DOC_MESSAGE
        doc = super().activity_gated_event(doc)

        descriptor_name = self.descriptors[doc["descriptor"]].get("name")
        if descriptor_name == ZOCALO_READING_PLAN_NAME:
            self._handle_zocalo_read_event(doc)
        elif descriptor_name == DocDescriptorNames.OAV_GRID_SNAPSHOT_TRIGGERED:
            scan_data_infos = self._handle_oav_grid_snapshot_triggered(doc)
            self.ispyb_ids = self.ispyb.update_deposition(
                self.ispyb_ids, scan_data_infos
            )
        self.ispyb.update_data_collection_group_table(
            self.data_collection_group_info, self.ispyb_ids.data_collection_group_id
        )

        return doc

    def _handle_zocalo_read_event(self, doc):
        assert self.data_collection_group_info, ASSERT_START_BEFORE_EVENT_DOC_MESSAGE
        crystal_summary = ""
        if self._processing_start_time is not None:
            proc_time = time() - self._processing_start_time
            crystal_summary = f"Zocalo processing took {proc_time:.2f} s. "
        bboxes: list[np.ndarray] = []
        ISPYB_ZOCALO_CALLBACK_LOGGER.info(
            f"Amending comment based on Zocalo reading doc: {format_doc_for_log(doc)}"
        )

        raw_results = get_processing_results_from_event("zocalo", doc)
        if len(raw_results) > 0:
            for n, res in enumerate(raw_results):
                bb = res["bounding_box"]
                diff = np.array(bb[1]) - np.array(bb[0])
                bboxes.append(diff)

                nicely_formatted_com = [
                    f"{np.round(com, 2)}" for com in res["centre_of_mass"]
                ]
                crystal_summary += (
                    f"Crystal {n + 1}: "
                    f"Strength {res['total_count']}; "
                    f"Position (grid boxes) {nicely_formatted_com}; "
                    f"Size (grid boxes) {bboxes[n]}; "
                )
        else:
            crystal_summary += "Zocalo found no crystals in this gridscan."
        assert self.ispyb_ids.data_collection_ids, (
            "No data collection to add results to"
        )

        self.data_collection_group_info.comments = (
            self.data_collection_group_info.comments or ""
        ) + crystal_summary

        self.ispyb.append_to_comment(
            self.ispyb_ids.data_collection_ids[0], crystal_summary
        )

    def _handle_oav_grid_snapshot_triggered(self, doc) -> Sequence[ScanDataInfo]:
        assert self.ispyb_ids.data_collection_ids, "No current data collection"
        assert self.params, "ISPyB handler didn't receive parameters!"
        assert self.data_collection_group_info, "No data collection group"
        data = doc["data"]
        data_collection_id = None
        data_collection_info = DataCollectionInfo(
            xtal_snapshot1=data.get("oav-grid_snapshot-last_path_full_overlay"),
            xtal_snapshot2=data.get("oav-grid_snapshot-last_path_outer"),
            xtal_snapshot3=data.get("oav-grid_snapshot-last_saved_path"),
            n_images=(
                data["oav-grid_snapshot-num_boxes_x"]
                * data["oav-grid_snapshot-num_boxes_y"]
            ),
        )
        microns_per_pixel_x = data["oav-microns_per_pixel_x"]
        microns_per_pixel_y = data["oav-microns_per_pixel_y"]
        data_collection_grid_info = DataCollectionGridInfo(
            dx_in_mm=data["oav-grid_snapshot-box_width"] * microns_per_pixel_x / 1000,
            dy_in_mm=data["oav-grid_snapshot-box_width"] * microns_per_pixel_y / 1000,
            steps_x=data["oav-grid_snapshot-num_boxes_x"],
            steps_y=data["oav-grid_snapshot-num_boxes_y"],
            microns_per_pixel_x=microns_per_pixel_x,
            microns_per_pixel_y=microns_per_pixel_y,
            snapshot_offset_x_pixel=int(data["oav-grid_snapshot-top_left_x"]),
            snapshot_offset_y_pixel=int(data["oav-grid_snapshot-top_left_y"]),
            orientation=Orientation.HORIZONTAL,
            snaked=True,
        )
        data_collection_info.comments = construct_comment_for_gridscan(
            data_collection_grid_info
        )

        if self.data_collection_group_info.comments:
            self.data_collection_group_info.comments += (
                f"by {data_collection_grid_info.steps_y}."
            )
        else:
            self.data_collection_group_info.comments = (
                f"Diffraction grid scan of "
                f"{data_collection_grid_info.steps_x} "
                f"by {data_collection_grid_info.steps_y} "
            )

        if len(self.ispyb_ids.data_collection_ids) > self._oav_snapshot_event_idx:
            data_collection_id = self.ispyb_ids.data_collection_ids[
                self._oav_snapshot_event_idx
            ]
        self._populate_axis_info(data_collection_info, doc["data"]["smargon-omega"])

        scan_data_info = ScanDataInfo(
            data_collection_info=data_collection_info,
            data_collection_id=data_collection_id,
            data_collection_grid_info=data_collection_grid_info,
        )
        ISPYB_ZOCALO_CALLBACK_LOGGER.info(
            "Updating ispyb data collection after oav snapshot."
        )
        self._oav_snapshot_event_idx += 1
        return [scan_data_info]

    def _populate_axis_info(
        self, data_collection_info: DataCollectionInfo, omega_start: float | None
    ):
        if omega_start is not None:
            omega_in_gda_space = -omega_start
            data_collection_info.omega_start = omega_in_gda_space
            data_collection_info.axis_start = omega_in_gda_space
            data_collection_info.axis_end = omega_in_gda_space
            data_collection_info.axis_range = 0

    def populate_info_for_update(
        self,
        event_sourced_data_collection_info: DataCollectionInfo,
        event_sourced_position_info: DataCollectionPositionInfo | None,
        params: DiffractionExperimentWithSample,
    ) -> Sequence[ScanDataInfo]:
        assert self.ispyb_ids.data_collection_ids, (
            "Expect at least one valid data collection to record scan data"
        )
        xy_scan_data_info = ScanDataInfo(
            data_collection_info=event_sourced_data_collection_info,
            data_collection_id=self.ispyb_ids.data_collection_ids[0],
        )
        scan_data_infos = [xy_scan_data_info]

        data_collection_id = (
            self.ispyb_ids.data_collection_ids[1]
            if len(self.ispyb_ids.data_collection_ids) > 1
            else None
        )
        xz_scan_data_info = ScanDataInfo(
            data_collection_info=event_sourced_data_collection_info,
            data_collection_id=data_collection_id,
        )
        scan_data_infos.append(xz_scan_data_info)
        return scan_data_infos

    def activity_gated_stop(self, doc: RunStop) -> RunStop:
        assert self.data_collection_group_info, (
            f"No data collection group info - stop document has been emitted before a {PlanNameConstants.GRID_DETECT_AND_DO_GRIDSCAN} start document"
        )
        if doc.get("run_start") == self._start_of_fgs_uid:
            self._processing_start_time = time()
        if doc.get("run_start") == self.uid_to_finalize_on:
            ISPYB_ZOCALO_CALLBACK_LOGGER.info(
                "ISPyB callback received stop document corresponding to start document "
                f"with uid: {self.uid_to_finalize_on}."
            )
            if self.ispyb_ids == IspybIds():
                raise ISPyBDepositionNotMade("ispyb was not initialised at run start")
            exception_type, message = SampleException.type_and_message_from_reason(
                doc.get("reason", "")
            )
            if exception_type:
                doc["reason"] = message
                self.data_collection_group_info.comments = message
                self.ispyb.update_data_collection_group_table(
                    self.data_collection_group_info,
                    self.ispyb_ids.data_collection_group_id,
                )
            self.data_collection_group_info = None
            return super().activity_gated_stop(doc)
        return self._tag_doc(doc)



import atexit
import json
import threading
from collections.abc import Callable
from dataclasses import asdict
from queue import Queue
from sys import argv
from traceback import format_exception
from typing import Any

from blueapi.core import BlueskyContext
from bluesky.callbacks.zmq import Publisher
from bluesky.run_engine import RunEngine
from bluesky.utils import MsgGenerator
from flask import Flask, request
from flask_restful import Api, Resource
from pydantic.dataclasses import dataclass

from mx_bluesky.common.external_interaction.callbacks.common.log_uid_tag_callback import (
    LogUidTaggingCallback,
)
from mx_bluesky.common.external_interaction.callbacks.common.logging_callback import (
    VerbosePlanExecutionLoggingCallback,
)
from mx_bluesky.common.parameters.components import MxBlueskyParameters
from mx_bluesky.common.parameters.constants import Actions, Status
from mx_bluesky.common.utils.exceptions import WarningException
from mx_bluesky.common.utils.log import (
    LOGGER,
    do_default_logging_setup,
    flush_debug_handler,
)
from mx_bluesky.common.utils.tracing import TRACER
from mx_bluesky.hyperion.experiment_plans.experiment_registry import (
    PLAN_REGISTRY,
    PlanNotFound,
)
from mx_bluesky.hyperion.external_interaction.agamemnon import (
    compare_params,
    update_params_from_agamemnon,
)
from mx_bluesky.hyperion.parameters.cli import parse_cli_args
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.utils.context import setup_context

VERBOSE_EVENT_LOGGING: bool | None = None


@dataclass
class Command:
    action: Actions
    devices: Any | None = None
    experiment: Callable[[Any, Any], MsgGenerator] | None = None
    parameters: MxBlueskyParameters | None = None


@dataclass
class StatusAndMessage:
    status: str
    message: str = ""

    def __init__(self, status: Status, message: str = "") -> None:
        self.status = status.value
        self.message = message


@dataclass
class ErrorStatusAndMessage(StatusAndMessage):
    exception_type: str = ""


def make_error_status_and_message(exception: Exception):
    return ErrorStatusAndMessage(
        status=Status.FAILED.value,
        message=repr(exception),
        exception_type=type(exception).__name__,
    )


class BlueskyRunner:
    def __init__(
        self,
        RE: RunEngine,
        context: BlueskyContext,
        skip_startup_connection=False,
    ) -> None:
        self.command_queue: Queue[Command] = Queue()
        self.current_status: StatusAndMessage = StatusAndMessage(Status.IDLE)
        self.last_run_aborted: bool = False
        self.logging_uid_tag_callback = LogUidTaggingCallback()
        self.context: BlueskyContext

        self.RE = RE
        self.context = context
        RE.subscribe(self.logging_uid_tag_callback)

        LOGGER.info("Connecting to external callback ZMQ proxy...")
        self.publisher = Publisher(f"localhost:{CONST.CALLBACK_0MQ_PROXY_PORTS[0]}")
        RE.subscribe(self.publisher)

        if VERBOSE_EVENT_LOGGING:
            RE.subscribe(VerbosePlanExecutionLoggingCallback())

        self.skip_startup_connection = skip_startup_connection
        if not self.skip_startup_connection:
            LOGGER.info("Initialising dodal devices...")
            for plan_name in PLAN_REGISTRY:
                PLAN_REGISTRY[plan_name]["setup"](context)

    def start(
        self,
        experiment: Callable,
        parameters: MxBlueskyParameters,
        plan_name: str,
    ) -> StatusAndMessage:
        LOGGER.info(f"Started with parameters: {parameters.model_dump_json(indent=2)}")

        devices: Any = PLAN_REGISTRY[plan_name]["setup"](self.context)

        if (
            self.current_status.status == Status.BUSY.value
            or self.current_status.status == Status.ABORTING.value
        ):
            return StatusAndMessage(Status.FAILED, "Bluesky already running")
        else:
            self.current_status = StatusAndMessage(Status.BUSY)
            self.command_queue.put(
                Command(
                    action=Actions.START,
                    devices=devices,
                    experiment=experiment,
                    parameters=parameters,
                )
            )
            return StatusAndMessage(Status.SUCCESS)

    def stopping_thread(self):
        try:
            self.RE.abort()
            self.current_status = StatusAndMessage(Status.IDLE)
        except Exception as e:
            self.current_status = make_error_status_and_message(e)

    def stop(self) -> StatusAndMessage:
        if self.current_status.status == Status.IDLE.value:
            return StatusAndMessage(Status.FAILED, "Bluesky not running")
        elif self.current_status.status == Status.ABORTING.value:
            return StatusAndMessage(Status.FAILED, "Bluesky already stopping")
        else:
            self.current_status = StatusAndMessage(Status.ABORTING)
            stopping_thread = threading.Thread(target=self.stopping_thread)
            stopping_thread.start()
            self.last_run_aborted = True
            return StatusAndMessage(Status.ABORTING)

    def shutdown(self):
        """Stops the run engine and the loop waiting for messages."""
        print("Shutting down: Stopping the run engine gracefully")
        self.stop()
        self.command_queue.put(Command(action=Actions.SHUTDOWN))

    def wait_on_queue(self):
        while True:
            command = self.command_queue.get()
            if command.action == Actions.SHUTDOWN:
                return
            elif command.action == Actions.START:
                if command.experiment is None:
                    raise ValueError("No experiment provided for START")
                try:
                    with TRACER.start_span("do_run"):
                        self.RE(command.experiment(command.devices, command.parameters))

                    self.current_status = StatusAndMessage(Status.IDLE)

                    self.last_run_aborted = False
                except WarningException as exception:
                    LOGGER.warning("Warning Exception", exc_info=True)
                    self.current_status = make_error_status_and_message(exception)
                except Exception as exception:
                    LOGGER.error("Exception on running plan", exc_info=True)

                    if self.last_run_aborted:
                        # Aborting will cause an exception here that we want to swallow
                        self.last_run_aborted = False
                    else:
                        self.current_status = make_error_status_and_message(exception)


def compose_start_args(context: BlueskyContext, plan_name: str, action: Actions):
    experiment_registry_entry = PLAN_REGISTRY.get(plan_name)
    if experiment_registry_entry is None:
        raise PlanNotFound(f"Experiment plan '{plan_name}' not found in registry.")

    experiment_internal_param_type = experiment_registry_entry.get("param_type")
    plan = context.plan_functions.get(plan_name)
    if experiment_internal_param_type is None:
        raise PlanNotFound(
            f"Corresponding internal param type for '{plan_name}' not found in registry."
        )
    if plan is None:
        raise PlanNotFound(
            f"Experiment plan '{plan_name}' not found in context. Context has {context.plan_functions.keys()}"
        )
    try:
        parameters = experiment_internal_param_type(**json.loads(request.data))
        parameters = update_params_from_agamemnon(parameters)
        compare_params(parameters)
        if parameters.model_extra:
            raise ValueError(f"Extra fields not allowed {parameters.model_extra}")
    except Exception as e:
        raise ValueError(
            f"Supplied parameters don't match the plan for this endpoint {request.data}, for plan {plan_name}"
        ) from e
    return plan, parameters, plan_name


class RunExperiment(Resource):
    def __init__(self, runner: BlueskyRunner, context: BlueskyContext) -> None:
        super().__init__()
        self.runner = runner
        self.context = context

    def put(self, plan_name: str, action: Actions):
        status_and_message = StatusAndMessage(Status.FAILED, f"{action} not understood")
        if action == Actions.START.value:
            try:
                plan, params, plan_name = compose_start_args(
                    self.context, plan_name, action
                )
                status_and_message = self.runner.start(plan, params, plan_name)
            except Exception as e:
                status_and_message = make_error_status_and_message(e)
                LOGGER.error(format_exception(e))

        elif action == Actions.STOP.value:
            status_and_message = self.runner.stop()
        # no idea why mypy gives an attribute error here but nowhere else for this
        # exact same situation...
        return asdict(status_and_message)  # type: ignore


class StopOrStatus(Resource):
    def __init__(self, runner: BlueskyRunner) -> None:
        super().__init__()
        self.runner: BlueskyRunner = runner

    def put(self, action):
        status_and_message = StatusAndMessage(Status.FAILED, f"{action} not understood")
        if action == Actions.STOP.value:
            status_and_message = self.runner.stop()
        return asdict(status_and_message)

    def get(self, **kwargs):
        action = kwargs.get("action")
        status_and_message = StatusAndMessage(Status.FAILED, f"{action} not understood")
        if action == Actions.STATUS.value:
            LOGGER.debug(
                f"Runner received status request - state of the runner object is: {self.runner.__dict__} - state of the RE is: {self.runner.RE.__dict__}"
            )
            status_and_message = self.runner.current_status
        return asdict(status_and_message)


class FlushLogs(Resource):
    def put(self, **kwargs):
        try:
            status_and_message = StatusAndMessage(
                Status.SUCCESS, f"Flushed debug log to {flush_debug_handler()}"
            )
        except Exception as e:
            status_and_message = StatusAndMessage(
                Status.FAILED, f"Failed to flush debug log: {e}"
            )
        return asdict(status_and_message)


def create_app(
    test_config=None,
    RE: RunEngine = RunEngine({}),
    skip_startup_connection: bool = False,
) -> tuple[Flask, BlueskyRunner]:
    context = setup_context(
        wait_for_connection=not skip_startup_connection,
    )
    runner = BlueskyRunner(
        RE,
        context=context,
        skip_startup_connection=skip_startup_connection,
    )
    app = Flask(__name__)
    if test_config:
        app.config.update(test_config)
    api = Api(app)
    api.add_resource(
        RunExperiment,
        "/<string:plan_name>/<string:action>",
        resource_class_args=[runner, context],
    )
    api.add_resource(
        FlushLogs,
        "/flush_debug_log",
    )
    api.add_resource(
        StopOrStatus,
        "/<string:action>",
        resource_class_args=[runner],
    )
    return app, runner


def create_targets():
    hyperion_port = 5005
    args = parse_cli_args()
    do_default_logging_setup(
        CONST.LOG_FILE_NAME, CONST.GRAYLOG_PORT, dev_mode=args.dev_mode
    )
    LOGGER.info(f"Hyperion launched with args:{argv}")
    app, runner = create_app(
        skip_startup_connection=args.skip_startup_connection,
    )
    return app, runner, hyperion_port, args.dev_mode


def main():
    app, runner, port, dev_mode = create_targets()
    atexit.register(runner.shutdown)
    flask_thread = threading.Thread(
        target=lambda: app.run(
            host="0.0.0.0", port=port, debug=True, use_reloader=False
        ),
        daemon=True,
    )
    flask_thread.start()
    LOGGER.info(f"Hyperion now listening on {port} ({'IN DEV' if dev_mode else ''})")
    runner.wait_on_queue()
    flask_thread.join()


if __name__ == "__main__":
    main()
